import typing

from inngest._internal import const, types

from .consts import ServerKind


class Capabilities(types.BaseModel):
    connect: str = "v1"
    in_band_sync: str = "v1"
    trust_probe: str = "v1"


class UnauthenticatedInspection(types.BaseModel):
    schema_version: str = "2024-05-24"

    function_count: int
    has_event_key: bool
    has_signing_key: bool
    has_signing_key_fallback: bool
    mode: ServerKind


class AuthenticatedInspection(types.BaseModel):
    schema_version: str = "2024-05-24"

    api_origin: str
    app_id: str
    authentication_succeeded: typing.Literal[True] = True
    capabilities: Capabilities = Capabilities()
    env: str | None
    event_api_origin: str
    event_key_hash: str | None
    framework: str
    function_count: int
    has_event_key: bool
    has_signing_key: bool
    has_signing_key_fallback: bool
    mode: ServerKind
    sdk_language: str = const.LANGUAGE
    sdk_version: str = const.VERSION
    serve_origin: str | None
    serve_path: str | None
    signing_key_fallback_hash: str | None
    signing_key_hash: str | None



from __future__ import annotations

import typing

from inngest._internal import types

from .models import (
    CallResult,
    Context,
    ContextSync,
    ReportedStep,
    ReportedStepSync,
)

if typing.TYPE_CHECKING:
    from inngest._internal import (
        client_lib,
        execution_lib,
        function,
        server_lib,
        step_lib,
    )


class BaseExecution(typing.Protocol):
    version: str
    _request: server_lib.ServerRequest

    async def report_step(
        self,
        step_info: step_lib.StepInfo,
    ) -> ReportedStep: ...

    async def run(
        self,
        client: client_lib.Inngest,
        ctx: Context,
        handler: execution_lib.FunctionHandlerAsync[typing.Any],
        fn: function.Function[typing.Any],
        output_type: object = types.EmptySentinel,
    ) -> CallResult: ...


class BaseExecutionSync(typing.Protocol):
    version: str
    _request: server_lib.ServerRequest

    def report_step(
        self,
        step_info: step_lib.StepInfo,
    ) -> ReportedStepSync: ...

    def run(
        self,
        client: client_lib.Inngest,
        ctx: ContextSync,
        handler: execution_lib.FunctionHandlerSync[typing.Any],
        fn: function.Function[typing.Any],
        output_type: object = types.EmptySentinel,
    ) -> CallResult: ...



"""Flask integration for Inngest."""

import json
import typing

import flask

from inngest._internal import (
    client_lib,
    comm_lib,
    config_lib,
    const,
    function,
    server_lib,
)

FRAMEWORK = server_lib.Framework.FLASK


def serve(
    app: flask.Flask,
    client: client_lib.Inngest,
    functions: list[function.Function[typing.Any]],
    *,
    public_path: str | None = None,
    serve_origin: str | None = None,
    serve_path: str | None = None,
) -> None:
    """
    Serve Inngest functions in a Flask app.

    Args:
    ----
        app: Flask app.
        client: Inngest client.
        functions: List of functions to serve.
        public_path: Path that the Inngest server sends requests to. This is only necessary if the SDK is behind a proxy that rewrites the path.
        serve_origin: Origin for serving Inngest functions. This is typically only useful during Docker-based development.
        serve_path: Path for serving Inngest functions. This is only useful if you don't want serve Inngest at the /api/inngest path.
    """

    handler = comm_lib.CommHandler(
        client=client,
        framework=FRAMEWORK,
        functions=functions,
        streaming=const.Streaming.DISABLE,  # Not supported yet.
    )

    async_mode = any(
        function.is_handler_async or function.is_on_failure_handler_async
        for function in functions
    )
    if async_mode:
        _create_handler_async(
            app,
            client,
            handler,
            public_path=public_path,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )
    else:
        _create_handler_sync(
            app,
            client,
            handler,
            public_path=public_path,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )


def _create_handler_async(
    app: flask.Flask,
    client: client_lib.Inngest,
    handler: comm_lib.CommHandler,
    *,
    public_path: str | None,
    serve_origin: str | None,
    serve_path: str | None,
) -> None:
    @app.route(
        config_lib.get_serve_path(serve_path) or const.DEFAULT_SERVE_PATH,
        methods=["GET", "POST", "PUT"],
    )
    async def inngest_api() -> flask.Response | str:
        comm_req = comm_lib.CommRequest(
            body=_get_body_bytes(),
            headers=dict(flask.request.headers.items()),
            public_path=public_path,
            query_params=flask.request.args,
            raw_request=flask.request,
            request_url=flask.request.url,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        if flask.request.method == "GET":
            return _to_response(
                client,
                handler.get_sync(comm_req),
            )

        if flask.request.method == "POST":
            return _to_response(
                client,
                await handler.post(comm_req),
            )

        if flask.request.method == "PUT":
            return _to_response(
                client,
                await handler.put(comm_req),
            )

        # Should be unreachable
        return ""


def _create_handler_sync(
    app: flask.Flask,
    client: client_lib.Inngest,
    handler: comm_lib.CommHandler,
    *,
    public_path: str | None,
    serve_origin: str | None,
    serve_path: str | None,
) -> None:
    @app.route(
        config_lib.get_serve_path(serve_path) or const.DEFAULT_SERVE_PATH,
        methods=["GET", "POST", "PUT"],
    )
    def inngest_api() -> flask.Response | str:
        comm_req = comm_lib.CommRequest(
            body=_get_body_bytes(),
            headers=dict(flask.request.headers.items()),
            public_path=public_path,
            query_params=flask.request.args,
            raw_request=flask.request,
            request_url=flask.request.url,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        if flask.request.method == "GET":
            return _to_response(
                client,
                handler.get_sync(comm_req),
            )

        if flask.request.method == "POST":
            return _to_response(
                client,
                handler.post_sync(comm_req),
            )

        if flask.request.method == "PUT":
            return _to_response(
                client,
                handler.put_sync(comm_req),
            )

        # Should be unreachable
        return ""


def _get_body_bytes() -> bytes:
    flask.request.get_data(as_text=True)
    return flask.request.data


def _to_response(
    client: client_lib.Inngest,
    comm_res: comm_lib.CommResponse,
) -> flask.Response:
    body = comm_res.body_bytes()
    if isinstance(body, Exception):
        comm_res = comm_lib.CommResponse.from_error(client.logger, body)
        body = json.dumps(comm_res.body).encode("utf-8")

    return flask.Response(
        headers=comm_res.headers,
        response=body,
        status=comm_res.status_code,
    )



import secrets
import string
import typing

import pydantic

import inngest

from .middleware import StateDriver


class _StatePlaceholder(pydantic.BaseModel):
    key: str


class InMemoryDriver(StateDriver):
    """
    In-memory driver for remote state middleware. This probably doesn't have any
    use besides being a reference.
    """

    # Marker to indicate that the data is stored remotely.
    _marker: typing.Final = "__REMOTE_STATE__"

    # Marker to indicate which strategy was used. This is useful for knowing
    # whether the official S3 driver was used.
    _strategy_marker: typing.Final = "__STRATEGY__"

    _strategy_identifier: typing.Final = "inngest/memory"

    def __init__(self) -> None:  # noqa: D107
        self._data: dict[str, object] = {}

    def _create_key(self) -> str:
        chars = string.ascii_letters + string.digits
        return "".join(secrets.choice(chars) for _ in range(32))

    def load_steps(self, steps: inngest.StepMemos) -> None:
        """
        Hydrate steps with remote state if necessary.
        """

        for step in steps.values():
            if not isinstance(step.data, dict):
                continue
            if self._marker not in step.data:
                continue
            if self._strategy_marker not in step.data:
                continue
            if step.data[self._strategy_marker] != self._strategy_identifier:
                continue

            placeholder = _StatePlaceholder.model_validate(step.data)

            step.data = self._data[placeholder.key]

    def save_step(
        self,
        run_id: str,
        value: object,
    ) -> dict[str, object]:
        """
        Save a step's output to the remote store and return a placeholder.
        """

        key = self._create_key()
        self._data[key] = value

        placeholder: dict[str, object] = {
            self._marker: True,
            self._strategy_marker: self._strategy_identifier,
            **_StatePlaceholder(key=key).model_dump(),
        }

        return placeholder



"""
Simulator for the DigitalOcean Function runtime.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to use it, we recommend copying this file into your source code.
"""

import dataclasses
import json
import typing

import flask

_PATH_PREFIX = "/api/v1/web"
_NAMESPACE = "fn-00000000"
_FUNCTION_NAME = "fake-name"
FULL_PATH = f"{_PATH_PREFIX}/{_NAMESPACE}/{_FUNCTION_NAME}"

_Main = typing.Callable[[typing.Any, typing.Any], typing.Any]


@dataclasses.dataclass
class _Context:
    api_host: str
    function_name: str


class DigitalOceanSimulator:
    """
    Simulator for the DigitalOcean Function runtime. This is necessary because
    DigitalOcean Functions don't seem to have a local development environment.
    If that's wrong or changes, this class can likely be deleted.
    """

    def __init__(self, main: _Main) -> None:
        """
        Args:
        ----
            main: DigitalOcean Function handler.
        """
        self.app = _create_app(main)


def _create_app(main: _Main) -> flask.Flask:
    app = flask.Flask(__name__)

    def handler() -> flask.Response:
        event: dict[str, object] = {
            "http": {
                "body": flask.request.data.decode("utf-8"),
                "headers": dict(flask.request.headers.items()),
                "method": flask.request.method,
                "path": "",
                "queryString": flask.request.query_string.decode("utf-8"),
            },
        }

        res = main(
            event,
            _Context(
                api_host=flask.request.url_root,
                function_name=f"/{_NAMESPACE}/{_FUNCTION_NAME}",
            ),
        )
        if not isinstance(res, dict):
            raise ValueError("response must be a dict")

        body = None
        if "body" in res:
            body = json.dumps(res["body"])

        headers = res.get("headers", {})
        if not isinstance(headers, dict):
            raise ValueError("headers must be a dict")

        status_code = res.get("statusCode", 200)
        if not isinstance(status_code, int):
            raise ValueError("statusCode must be an int")

        return flask.Response(
            headers=headers,
            response=body,
            status=status_code,
        )

    @app.route("/", methods=["GET", "POST", "PUT"])
    def root_handler() -> flask.Response:
        return handler()

    @app.route(FULL_PATH, methods=["GET", "POST", "PUT"])
    def full_path_handler() -> flask.Response:
        return handler()

    return app



from __future__ import annotations

import typing

from inngest._internal import (
    client_lib,
    execution_lib,
    function,
    log,
    step_lib,
)

from .middleware import MiddlewareSync


class LoggerMiddleware(MiddlewareSync):
    def __init__(self, client: client_lib.Inngest, raw_request: object) -> None:
        super().__init__(client, raw_request)
        # Start with logging disabled (during step replay)
        log.disable_logging()
        self._filtered_logger: log.FilteredLogger | None = None

    def before_execution(self) -> None:
        # Enable logging when actual execution begins
        log.enable_logging()

    def transform_input(
        self,
        ctx: execution_lib.Context | execution_lib.ContextSync,
        function: function.Function[typing.Any],
        steps: step_lib.StepMemos,
    ) -> None:
        if self._filtered_logger is None:
            self._filtered_logger = log.FilteredLogger(ctx.logger)
        ctx.logger = self._filtered_logger  # type: ignore



from __future__ import annotations

import asyncio
import concurrent.futures
import functools
import http
import os
import typing
import urllib.parse

import httpx

from inngest._internal import (
    client_lib,
    const,
    env_lib,
    errors,
    execution_lib,
    function,
    middleware_lib,
    net,
    server_lib,
    step_lib,
    transforms,
    types,
)

from .models import CommRequest, CommResponse
from .utils import parse_query_params, wrap_handler, wrap_handler_sync


class CommHandler:
    _base_url: str
    _client: client_lib.Inngest
    _fns: dict[str, function.Function[typing.Any]]
    _framework: server_lib.Framework
    _mode: server_lib.ServerKind
    _signing_key: str | None
    _signing_key_fallback: str | None

    def __init__(
        self,
        *,
        client: client_lib.Inngest,
        framework: server_lib.Framework,
        functions: list[function.Function[typing.Any]],
        streaming: const.Streaming | None,
    ) -> None:
        # In-band syncing is opt-out.
        self._allow_in_band_sync = not env_lib.is_false(
            const.EnvKey.ALLOW_IN_BAND_SYNC,
        )

        self._client = client
        self._http_client = client._http_client
        self._mode = client._mode
        self._api_origin = client.api_origin
        self._fns = {fn.get_id(): fn for fn in functions}
        self._framework = framework

        if streaming is None:
            streaming = env_lib.get_streaming(const.EnvKey.STREAMING)
        self._streaming = streaming or const.Streaming.DISABLE
        if self._streaming == const.Streaming.FORCE:
            self._client.logger.warning("Streaming responses are enabled")

        # TODO: Graduate this to a config option, rather than an env var.
        thread_pool_max_workers = env_lib.get_int(
            const.EnvKey.THREAD_POOL_MAX_WORKERS,
        )
        if thread_pool_max_workers == 0:
            self._client.logger.debug(
                "Skipping thread pool creation because max workers is 0",
            )
            self._thread_pool = None
        else:
            # We need a thread pool when both of the following are true:
            # 1. CommHandler is called from an async context (e.g. using FastAPI
            #   or Connect).
            # 2. Executing a non-async function.
            #
            # When the aforementioned situation happens, we need a thread pool
            # to run the function in a non-blocking way. Without a thread pool,
            # blocking operations will block the event loop.
            #
            # We don't need the thread pool when CommHandler is called from a
            # non-async context because we can assume that the HTTP framework
            # (e.g.  Flask) created a thread for the request.
            self._thread_pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=thread_pool_max_workers,
            )

        signing_key = client.signing_key
        if signing_key is None:
            if self._client.is_production:
                signing_key = os.getenv(const.EnvKey.SIGNING_KEY.value)
                if signing_key is None:
                    self._client.logger.error("missing signing key")
                    raise errors.SigningKeyMissingError()
        self._signing_key = signing_key

        self._signing_key_fallback = client.signing_key_fallback

    @wrap_handler()
    async def post(
        self,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> CommResponse | Exception:
        params = parse_query_params(req.query_params)
        if isinstance(params, Exception):
            return params

        if params.probe is server_lib.Probe.TRUST:
            return CommResponse()

        server_kind = transforms.get_server_kind(req.headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        middleware = middleware_lib.MiddlewareManager.from_client(
            self._client,
            req.raw_request,
            req.timings,
        )

        request = server_lib.ServerRequest.from_raw(req.body)
        if isinstance(request, Exception):
            return request

        if params.fn_id is None:
            return errors.QueryParamMissingError(
                server_lib.QueryParamKey.FUNCTION_ID.value
            )

        # Get the function we should call.
        fn = self._get_function(params.fn_id)
        if isinstance(fn, Exception):
            return fn

        events = request.events
        steps = request.steps
        if request.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            with req.timings.use_api:
                fetched_events, fetched_steps = await asyncio.gather(
                    self._client._get_batch(request.ctx.run_id),
                    self._client._get_steps(request.ctx.run_id),
                )
                if isinstance(fetched_events, Exception):
                    return fetched_events
                events = fetched_events
                if isinstance(fetched_steps, Exception):
                    return fetched_steps
                steps = fetched_steps
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return Exception("events not in request")

        memos = step_lib.StepMemos.from_raw(steps)

        if fn.is_handler_async:
            # Don't await because we might need to stream the response.
            call_res_task = asyncio.create_task(
                fn.call(
                    self._client,
                    execution_lib.Context(
                        attempt=request.ctx.attempt,
                        event=request.event,
                        events=events,
                        group=step_lib.Group(),
                        logger=self._client.logger,
                        run_id=request.ctx.run_id,
                        step=step_lib.Step(
                            self._client,
                            execution_lib.ExecutionV0(
                                memos,
                                middleware,
                                request,
                                params.step_id,
                                req.timings,
                            ),
                            middleware,
                            step_lib.StepIDCounter(),
                            params.step_id,
                        ),
                    ),
                    params.fn_id,
                    middleware,
                )
            )

            if self._streaming is const.Streaming.FORCE:
                return CommResponse.create_streaming(
                    self._client.logger,
                    call_res_task,
                    self._client.env,
                    self._framework,
                    server_kind,
                    req.timings,
                )

            call_res = await call_res_task
        else:
            fn_call = functools.partial(
                fn.call_sync,
                self._client,
                execution_lib.ContextSync(
                    attempt=request.ctx.attempt,
                    event=request.event,
                    events=events,
                    group=step_lib.GroupSync(),
                    logger=self._client.logger,
                    run_id=request.ctx.run_id,
                    step=step_lib.StepSync(
                        self._client,
                        execution_lib.ExecutionV0Sync(
                            memos,
                            middleware,
                            request,
                            params.step_id,
                            req.timings,
                        ),
                        middleware,
                        step_lib.StepIDCounter(),
                        params.step_id,
                    ),
                ),
                params.fn_id,
                middleware,
            )

            if self._thread_pool is not None:
                loop = asyncio.get_running_loop()
                call_res = await loop.run_in_executor(
                    self._thread_pool,
                    fn_call,
                )
            else:
                call_res = fn_call()

        return CommResponse.from_call_result(
            self._client.logger,
            call_res,
            self._client.env,
            self._framework,
            server_kind,
        )

    @wrap_handler_sync()
    def post_sync(
        self,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> CommResponse | Exception:
        params = parse_query_params(req.query_params)
        if isinstance(params, Exception):
            return params

        if params.probe is server_lib.Probe.TRUST:
            return CommResponse()

        server_kind = transforms.get_server_kind(req.headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        middleware = middleware_lib.MiddlewareManager.from_client(
            self._client,
            req.raw_request,
            req.timings,
        )

        request = server_lib.ServerRequest.from_raw(req.body)
        if isinstance(request, Exception):
            return request

        if params.fn_id is None:
            return errors.QueryParamMissingError(
                server_lib.QueryParamKey.FUNCTION_ID.value
            )

        # Get the function we should call.
        fn = self._get_function(params.fn_id)
        if isinstance(fn, Exception):
            return fn

        events = request.events
        steps = request.steps
        if request.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            with req.timings.use_api:
                fetched_events = self._client._get_batch_sync(
                    request.ctx.run_id
                )
                if isinstance(fetched_events, Exception):
                    return fetched_events
                events = fetched_events

                fetched_steps = self._client._get_steps_sync(request.ctx.run_id)
                if isinstance(fetched_steps, Exception):
                    return fetched_steps
                steps = fetched_steps
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return Exception("events not in request")

        memos = step_lib.StepMemos.from_raw(steps)

        call_res = fn.call_sync(
            self._client,
            execution_lib.ContextSync(
                attempt=request.ctx.attempt,
                event=request.event,
                events=events,
                group=step_lib.GroupSync(),
                logger=self._client.logger,
                run_id=request.ctx.run_id,
                step=step_lib.StepSync(
                    self._client,
                    execution_lib.ExecutionV0Sync(
                        memos,
                        middleware,
                        request,
                        params.step_id,
                        req.timings,
                    ),
                    middleware,
                    step_lib.StepIDCounter(),
                    params.step_id,
                ),
            ),
            params.fn_id,
            middleware,
        )

        return CommResponse.from_call_result(
            self._client.logger,
            call_res,
            self._client.env,
            self._framework,
            server_kind,
        )

    def _get_function(
        self, fn_id: str
    ) -> types.MaybeError[function.Function[typing.Any]]:
        # Look for the function ID in the list of user functions, but also
        # look for it in the list of on_failure functions.
        for _fn in self._fns.values():
            if _fn.get_id() == fn_id:
                return _fn
            if _fn.on_failure_fn_id == fn_id:
                return _fn

        # If we didn't find the function ID, it might be because the function ID
        # in the request uses the old format that didn't include the app ID.
        # We'll prefix the function ID with the app ID and try again. This logic
        # can be deleted when no one is using Python SDK versions below 0.3.0
        # anymore.
        app_and_fn_id = f"{self._client.app_id}-{fn_id}"
        for _fn in self._fns.values():
            if _fn.get_id() == app_and_fn_id:
                return _fn
            if _fn.on_failure_fn_id == app_and_fn_id:
                return _fn

        return errors.FunctionNotFoundError(f"function {fn_id} not found")

    @wrap_handler_sync(require_signature=False)
    def get_sync(
        self,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> types.MaybeError[CommResponse]:
        """Handle Dev Server's auto-discovery."""

        server_kind = transforms.get_server_kind(req.headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        if server_kind is not None and server_kind != self._mode:
            # Tell Dev Server to leave the app alone since it's in production
            # mode.
            return CommResponse(
                body={},
                status_code=403,
            )

        inspection = _build_inspection_response(
            self,
            req,
            request_signing_key,
        )
        if isinstance(inspection, Exception):
            return inspection

        res_body = inspection.to_dict()
        if isinstance(res_body, Exception):
            return res_body

        return CommResponse(
            body=res_body,
            status_code=200,
        )

    @wrap_handler(require_signature=False)
    async def put(
        self: CommHandler,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> CommResponse | Exception:
        """Handle a PUT request."""

        self._client.logger.debug("Syncing app")
        syncer = Syncer(logger=self._client.logger)

        if (
            req.headers.get(server_lib.HeaderKey.SYNC_KIND.value)
            == server_lib.SyncKind.IN_BAND.value
            and self._allow_in_band_sync
        ):
            err: Exception | None = None
            if isinstance(request_signing_key, Exception):
                err = request_signing_key
            elif request_signing_key is None:
                err = Exception("request must be signed for in-band sync")
            if err is not None:
                return CommResponse.from_error(
                    self._client.logger,
                    err,
                    status=http.HTTPStatus.UNAUTHORIZED,
                )
            return syncer.in_band(self, req, request_signing_key)

        return await syncer.out_of_band(self, req)

    @wrap_handler_sync(require_signature=False)
    def put_sync(
        self: CommHandler,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> CommResponse | Exception:
        """Handle a PUT request."""

        self._client.logger.debug("Syncing app")
        syncer = Syncer(logger=self._client.logger)

        if (
            req.headers.get(server_lib.HeaderKey.SYNC_KIND.value)
            == server_lib.SyncKind.IN_BAND.value
            and self._allow_in_band_sync
        ):
            err: Exception | None = None
            if isinstance(request_signing_key, Exception):
                err = request_signing_key
            elif request_signing_key is None:
                err = Exception("request must be signed for in-band sync")
            if err is not None:
                return CommResponse.from_error(
                    self._client.logger,
                    err,
                    status=http.HTTPStatus.UNAUTHORIZED,
                )

            return syncer.in_band(self, req, request_signing_key)

        return syncer.out_of_band_sync(self, req)


def _build_inspection_response(
    handler: CommHandler,
    req: CommRequest,
    request_signing_key: types.MaybeError[str | None],
) -> types.MaybeError[
    server_lib.AuthenticatedInspection | server_lib.UnauthenticatedInspection
]:
    server_kind = transforms.get_server_kind(req.headers)
    if isinstance(server_kind, Exception):
        handler._client.logger.error(server_kind)
        server_kind = None

    is_signed = isinstance(request_signing_key, str)
    if is_signed:
        event_key_hash = (
            transforms.hash_event_key(handler._client.event_key)
            if handler._client.event_key
            else None
        )

        signing_key_hash = (
            transforms.hash_signing_key(handler._signing_key)
            if handler._signing_key
            else None
        )

        signing_key_fallback_hash = (
            transforms.hash_signing_key(handler._signing_key_fallback)
            if handler._signing_key_fallback
            else None
        )

        return server_lib.AuthenticatedInspection(
            api_origin=handler._client.api_origin,
            app_id=handler._client.app_id,
            authentication_succeeded=True,
            env=handler._client.env,
            event_api_origin=handler._client.event_api_origin,
            event_key_hash=event_key_hash,
            framework=handler._framework.value,
            function_count=len(handler._fns),
            has_event_key=handler._client.event_key is not None,
            has_signing_key=handler._signing_key is not None,
            has_signing_key_fallback=handler._signing_key_fallback is not None,
            mode=handler._mode,
            serve_origin=req.serve_origin,
            serve_path=req.serve_path,
            signing_key_fallback_hash=signing_key_fallback_hash,
            signing_key_hash=signing_key_hash,
        )

    return server_lib.UnauthenticatedInspection(
        function_count=len(handler._fns),
        has_event_key=handler._client.event_key is not None,
        has_signing_key=handler._signing_key is not None,
        has_signing_key_fallback=handler._signing_key_fallback is not None,
        mode=handler._mode,
    )


class Syncer:
    def __init__(self, logger: types.Logger) -> None:
        self._logger = logger

    def in_band(
        self,
        handler: CommHandler,
        req: CommRequest,
        request_signing_key: types.MaybeError[str | None],
    ) -> types.MaybeError[CommResponse]:
        if not isinstance(request_signing_key, str):
            # This should be checked earlier, but we'll also check it here since
            # it's critical
            return Exception("request must be signed for in-band sync")

        req_body = server_lib.InBandSynchronizeRequest.from_raw(req.body)
        if isinstance(req_body, Exception):
            return req_body

        app_url = net.create_serve_url(
            public_path=req.public_path,
            request_url=req_body.url,
            serve_origin=req.serve_origin,
            serve_path=req.serve_path,
        )

        fn_configs = get_function_configs(app_url, handler._fns)
        if isinstance(fn_configs, Exception):
            return fn_configs

        inspection = _build_inspection_response(
            handler,
            req,
            request_signing_key,
        )
        if isinstance(inspection, Exception):
            return inspection
        if isinstance(inspection, server_lib.UnauthenticatedInspection):
            # Unreachable
            return Exception("request must be signed for in-band sync")

        res_body = server_lib.InBandSynchronizeResponse(
            app_id=handler._client.app_id,
            app_version=handler._client.app_version,
            env=handler._client.env,
            framework=handler._framework,
            functions=fn_configs,
            inspection=inspection,
            platform=None,
            url=app_url,
        ).to_dict()
        if isinstance(res_body, Exception):
            return res_body

        # Remove any None values from the response body. If we don't Go
        # marshalling may break in the Inngest server. Specifically, we saw this
        # with the concurrency scope.
        res_body = transforms.deep_strip_none(res_body)

        self._logger.debug("Responding to in-band sync")

        return CommResponse(
            body=res_body,
            headers={
                server_lib.HeaderKey.SYNC_KIND.value: server_lib.SyncKind.IN_BAND.value,
            },
        )

    def _create_out_of_band_request(
        self,
        handler: CommHandler,
        req: CommRequest,
    ) -> types.MaybeError[CommResponse | httpx.Request]:
        app_url = net.create_serve_url(
            public_path=req.public_path,
            request_url=req.request_url,
            serve_origin=req.serve_origin,
            serve_path=req.serve_path,
        )

        server_kind = transforms.get_server_kind(req.headers)
        if isinstance(server_kind, Exception):
            handler._client.logger.error(server_kind)
            server_kind = None

        if server_kind is not None and server_kind != handler._mode:
            msg: str
            if server_kind == server_lib.ServerKind.DEV_SERVER:
                msg = "Sync rejected since it's from a Dev Server but expected Cloud"
            else:
                msg = "Sync rejected since it's from Cloud but expected Dev Server"

            handler._client.logger.error(msg)
            return CommResponse.from_error_code(
                server_lib.ErrorCode.SERVER_KIND_MISMATCH,
                msg,
                http.HTTPStatus.BAD_REQUEST,
            )

        params = parse_query_params(req.query_params)
        if isinstance(params, Exception):
            return params

        registration_url = urllib.parse.urljoin(
            handler._api_origin,
            "/fn/register",
        )

        fn_configs = get_function_configs(app_url, handler._fns)
        if isinstance(fn_configs, Exception):
            return fn_configs

        body = server_lib.SynchronizeRequest(
            app_name=handler._client.app_id,
            app_version=handler._client.app_version,
            deploy_type=server_lib.DeployType.PING,
            framework=handler._framework,
            functions=fn_configs,
            sdk=f"{const.LANGUAGE}:v{const.VERSION}",
            url=app_url,
            v="0.1",
        ).to_dict()
        if isinstance(body, Exception):
            return body

        headers = net.create_headers(
            env=handler._client.env,
            framework=handler._framework,
            server_kind=server_kind,
        )

        outgoing_params = {}
        if params.sync_id is not None:
            outgoing_params[server_lib.QueryParamKey.SYNC_ID.value] = (
                params.sync_id
            )

        # Remove any None values from the response body. If we don't Go
        # marshalling may break in the Inngest server. Specifically, we saw this
        # with the concurrency scope.
        body = transforms.deep_strip_none(body)

        return handler._http_client.build_httpx_request(
            "POST",
            registration_url,
            headers=headers,
            json=body,
            params=outgoing_params,
            timeout=30,
        )

    def _parse_out_of_band_response(
        self,
        handler: CommHandler,
        res: httpx.Response,
    ) -> types.MaybeError[CommResponse]:
        try:
            server_res_body = res.json()
        except Exception:
            return errors.RegistrationFailedError("response is not valid JSON")

        if not isinstance(server_res_body, dict):
            return errors.RegistrationFailedError("response is not an object")

        if res.status_code >= 400:
            msg = server_res_body.get("error")
            if not isinstance(msg, str):
                msg = "registration failed"
            comm_res = CommResponse.from_error(
                handler._client.logger,
                errors.RegistrationFailedError(msg.strip()),
            )
            comm_res.status_code = res.status_code

        return CommResponse(
            body=server_res_body,
            headers={
                server_lib.HeaderKey.SYNC_KIND.value: server_lib.SyncKind.OUT_OF_BAND.value,
            },
        )

    async def out_of_band(
        self,
        handler: CommHandler,
        req: CommRequest,
    ) -> types.MaybeError[CommResponse]:
        prep = self._create_out_of_band_request(handler, req)
        if isinstance(prep, Exception):
            return prep
        if isinstance(prep, CommResponse):
            return prep

        self._logger.debug(f"Sending out-of-band sync request to {prep.url}")

        res = await net.fetch_with_auth_fallback(
            handler._http_client._http_client,
            handler._http_client._http_client_sync,
            prep,
            signing_key=handler._signing_key,
            signing_key_fallback=handler._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        return self._parse_out_of_band_response(handler, res)

    def out_of_band_sync(
        self,
        handler: CommHandler,
        req: CommRequest,
    ) -> types.MaybeError[CommResponse]:
        prep = self._create_out_of_band_request(handler, req)
        if isinstance(prep, Exception):
            return prep
        if isinstance(prep, CommResponse):
            return prep

        self._logger.debug(f"Sending out-of-band sync request to {prep.url}")

        res = net.fetch_with_auth_fallback_sync(
            handler._http_client._http_client_sync,
            prep,
            signing_key=handler._signing_key,
            signing_key_fallback=handler._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        return self._parse_out_of_band_response(handler, res)


def get_function_configs(
    app_url: str,
    fns: dict[str, function.Function[typing.Any]],
) -> types.MaybeError[list[server_lib.FunctionConfig]]:
    configs: list[server_lib.FunctionConfig] = []
    for fn in fns.values():
        config = fn.get_config(app_url)
        configs.append(config.main)

        if config.on_failure is not None:
            configs.append(config.on_failure)

    if len(configs) == 0:
        return errors.FunctionConfigInvalidError("no functions found")
    return configs



import datetime

from . import transforms


def test_hash_signing_key() -> None:
    signing_key = "signkey-prod-568c4116828e6e8384a554153722df93022e5cd29a6c2d1b0444a19a807ff315"
    expectation = (
        "2e64ca0edc850db32ff684f967822c828f99cf57862e43205fdcf2eff8d95180"
    )
    assert transforms.hash_signing_key(signing_key) == expectation


def test_to_duration_str() -> None:
    # Exact units
    assert transforms.to_duration_str(1000) == "1s"
    assert transforms.to_duration_str(datetime.timedelta(seconds=1)) == "1s"
    assert transforms.to_duration_str(datetime.timedelta(seconds=30)) == "30s"
    assert transforms.to_duration_str(datetime.timedelta(minutes=2)) == "2m"
    assert transforms.to_duration_str(datetime.timedelta(hours=2)) == "2h"
    assert transforms.to_duration_str(datetime.timedelta(days=3)) == "3d"
    assert transforms.to_duration_str(datetime.timedelta(weeks=2)) == "2w"

    # Non-exact-unit durations must not truncate
    assert transforms.to_duration_str(datetime.timedelta(seconds=90)) == "90s"
    assert transforms.to_duration_str(datetime.timedelta(minutes=90)) == "90m"
    assert transforms.to_duration_str(datetime.timedelta(hours=36)) == "36h"
    assert transforms.to_duration_str(datetime.timedelta(days=10)) == "10d"
    assert (
        transforms.to_duration_str(datetime.timedelta(minutes=2, seconds=30))
        == "150s"
    )

    # Sub-second durations are errors
    result = transforms.to_duration_str(datetime.timedelta(milliseconds=500))
    assert isinstance(result, Exception)



import contextvars
import typing

from inngest._internal import server_lib, types

from .base import ResponseInterrupt, SkipInterrupt, StepResponse

# Create a context variable to track if we're in a parallel group.
in_parallel = contextvars.ContextVar("in_parallel", default=False)


class Group:
    async def parallel(
        self,
        callables: tuple[typing.Callable[[], typing.Awaitable[types.T]], ...],
        parallel_mode: server_lib.ParallelMode = server_lib.ParallelMode.WAIT,
    ) -> tuple[types.T, ...]:
        """
        Run multiple steps in parallel.

        Args:
        ----
            callables: An arbitrary number of step callbacks to run. These are callables that contain the step (e.g. `lambda: step.run("my_step", my_step_fn)`.
            parallel_mode: Execution mode. Defaults to `ParallelMode.WAIT`
        """

        token = in_parallel.set(True)

        try:
            outputs = tuple[types.T]()
            responses: list[StepResponse] = []

            # Discover steps in callables.
            for cb in callables:
                try:
                    output = await cb()
                    outputs = (*outputs, output)
                except ResponseInterrupt as interrupt:
                    responses = [*responses, *interrupt.responses]
                except SkipInterrupt:
                    pass

            if len(responses) > 0:
                for r in responses:
                    r.step.set_parallel_mode(parallel_mode)
                raise ResponseInterrupt(responses)

            return outputs
        finally:
            # No longer tell steps that they're running in parallel.
            in_parallel.reset(token)


class GroupSync:
    def parallel(
        self,
        callables: tuple[typing.Callable[[], types.T], ...],
        parallel_mode: server_lib.ParallelMode = server_lib.ParallelMode.WAIT,
    ) -> tuple[types.T, ...]:
        """
        Run multiple steps in parallel in a synchronous context (e.g. not asyncio).

        Args:
        ----
            callables: An arbitrary number of step callbacks to run. These are callables that contain the step (e.g. `lambda: step.run("my_step", my_step_fn)`.
            parallel_mode: Execution mode. Defaults to `ParallelMode.WAIT`
        """

        # Tell steps that they're running in parallel.
        token = in_parallel.set(True)

        try:
            outputs = tuple[types.T]()
            responses: list[StepResponse] = []

            # Discover steps in callables.
            for cb in callables:
                try:
                    output = cb()
                    outputs = (*outputs, output)
                except ResponseInterrupt as interrupt:
                    responses = [*responses, *interrupt.responses]
                except SkipInterrupt:
                    pass

            if len(responses) > 0:
                for r in responses:
                    r.step.set_parallel_mode(parallel_mode)
                raise ResponseInterrupt(responses)

            return outputs
        finally:
            # No longer tell steps that they're running in parallel.
            in_parallel.reset(token)



import collections
import dataclasses
import time
import typing


@dataclasses.dataclass
class _BufferItem:
    data: bytes
    id: str
    timestamp: float


class SizeConstrainedBuffer:
    """
    Buffer for storing execution replies awaiting server acknowledgment.

    If the WebSocket connection drops before acknowledgment, these messages can
    be flushed via HTTP as a fallback.

    Features:
        - Maximum size enforcement: Oldest items are evicted when full
        - Timestamp tracking: Items can be retrieved by age for TTL-based flushing
        - O(1) add/delete/get operations using OrderedDict

    Used by _ExecutionHandler to implement reliable message delivery even
    across connection interruptions.
    """

    def __init__(
        self,
        max_size_bytes: int,
        on_evict: typing.Callable[[str], None] | None = None,
        on_reject: typing.Callable[[str], None] | None = None,
    ):
        """
        Initialize the buffer with a maximum size constraint.

        Args:
            max_size_bytes: Maximum total size of all items' data.
            on_evict: Existing item is evicted to make room for a new one.
            on_reject: New item is rejected because it exceeds the max buffer size by itself.
        """

        self._current_size = 0
        self._items: collections.OrderedDict[str, _BufferItem] = (
            collections.OrderedDict()
        )
        self._on_evict = on_evict
        self._on_reject = on_reject

        self._max_size_bytes = max_size_bytes
        if self._max_size_bytes <= 0:
            raise ValueError("max_size_bytes must be greater than 0")

    def add(self, item_id: str, data: bytes) -> bool:
        """
        Add item to buffer. If adding would exceed size limit, evicts oldest
        items until there is enough space. Returns True if item was added
        successfully. If the item is larger than the max size, it is not added
        to the buffer.
        """

        item_size = len(data)

        if item_size > self._max_size_bytes:
            if self._on_reject is not None:
                self._on_reject(item_id)
            return False

        # Remove existing item with same ID if it exists.
        if item_id in self._items:
            self._current_size -= len(self._items[item_id].data)
            del self._items[item_id]

        # If adding would exceed limit, evict oldest items until there's enough
        # space.
        while (
            self._current_size + item_size > self._max_size_bytes
            and self._items
        ):
            # Remove oldest item.
            _, oldest_item = self._items.popitem(last=False)
            self._current_size -= len(oldest_item.data)
            if self._on_evict is not None:
                self._on_evict(oldest_item.id)

        # Add new item.
        item = _BufferItem(
            data=data,
            id=item_id,
            timestamp=time.time(),
        )
        self._items[item_id] = item
        self._current_size += item_size
        return True

    def get(self, item_id: str) -> bytes | None:
        """
        Get item by ID without removing it.
        """

        item = self._items.get(item_id)
        if item is None:
            return None
        return item.data

    def delete(self, item_id: str) -> bool:
        """
        Delete item by ID. Returns True if item was found and deleted.
        """

        if item_id not in self._items:
            return False

        item = self._items[item_id]
        self._current_size -= len(item.data)
        del self._items[item_id]
        return True

    def get_older_than(self, seconds: float) -> list[tuple[str, bytes]]:
        """
        Get all items that were inserted at least `seconds` ago.
        """

        cutoff_time = time.time() - seconds
        result = []

        for item in self._items.values():
            if item.timestamp <= cutoff_time:
                result.append((item.id, item.data))

        return result



from __future__ import annotations

import asyncio
import dataclasses
import typing
import unittest.mock

import inngest
from inngest._internal import (
    async_lib,
    execution_lib,
    middleware_lib,
    net,
    server_lib,
    step_lib,
)

from .client import Inngest
from .consts import Status, Timeout
from .errors import UnstubbedStepError


def trigger(
    fn: inngest.Function[typing.Any],
    event: inngest.Event | list[inngest.Event],
    client: Inngest,
    *,
    step_stubs: dict[str, object] | None = None,
) -> _Result:
    """
    Trigger a function.

    Args:
    ----
        fn: Function to trigger.
        event: Triggering event.
        client: Mock Inngest client.
        step_stubs: Static step stubs. Keys are step IDs and values are stubs.
    """

    if not isinstance(event, list):
        event = [event]
    elif len(event) == 0:
        raise Exception("Must provide at least 1 event")

    if step_stubs is None:
        step_stubs = {}

    timings = net.ServerTimings()
    stack: list[str] = []
    steps: dict[str, object] = {}
    planned = set[str]()
    attempt = 0

    max_attempt = 4
    if fn._opts.retries is not None:
        max_attempt = fn._opts.retries

    while True:
        step_id: str | None = None
        if len(planned) > 0:
            step_id = planned.pop()

        logger = unittest.mock.Mock()
        request = server_lib.ServerRequest(
            ctx=server_lib.ServerRequestCtx(
                attempt=attempt,
                disable_immediate_execution=True,
                run_id="test",
                stack=server_lib.ServerRequestCtxStack(stack=stack),
            ),
            event=event[0],
            events=event,
            steps=steps,
            use_api=False,
        )
        middleware = middleware_lib.MiddlewareManager.from_client(
            client,
            {},
            timings,
        )

        memos = step_lib.StepMemos.from_raw(steps)

        ctx: execution_lib.Context | execution_lib.ContextSync
        if fn.is_handler_async:
            ctx = execution_lib.Context(
                attempt=request.ctx.attempt,
                event=event[0],
                events=event,
                group=step_lib.Group(),
                logger=logger,
                run_id=request.ctx.run_id,
                step=step_lib.Step(
                    client,
                    execution_lib.ExecutionV0(
                        memos,
                        middleware,
                        request,
                        step_id,
                        timings,
                    ),
                    middleware,
                    step_lib.StepIDCounter(),
                    step_id,
                ),
            )

            loop = async_lib.get_event_loop()
            if loop is None:
                loop = asyncio.new_event_loop()

            res = loop.run_until_complete(
                fn.call(
                    client,
                    ctx,
                    fn.id,
                    middleware,
                )
            )
        else:
            ctx = execution_lib.ContextSync(
                attempt=request.ctx.attempt,
                event=event[0],
                events=event,
                group=step_lib.GroupSync(),
                logger=logger,
                run_id=request.ctx.run_id,
                step=step_lib.StepSync(
                    client,
                    execution_lib.ExecutionV0Sync(
                        memos,
                        middleware,
                        request,
                        step_id,
                        timings,
                    ),
                    middleware,
                    step_lib.StepIDCounter(),
                    step_id,
                ),
            )

            res = fn.call_sync(
                client,
                ctx,
                fn.id,
                middleware,
            )

        if res.error:
            if attempt >= max_attempt:
                return _Result(
                    error=res.error,
                    output=None,
                    status=Status.FAILED,
                )

            attempt += 1
            continue

        if res.multi:
            for step in res.multi:
                if not step.step:
                    # Unreachable
                    continue

                if step.error:
                    if attempt >= max_attempt:
                        return _Result(
                            error=step.error,
                            output=None,
                            status=Status.FAILED,
                        )

                    attempt += 1
                    continue

                if step.step.display_name in step_stubs:
                    stub = step_stubs[step.step.display_name]
                    if stub is Timeout:
                        stub = None

                    steps[step.step.id] = stub
                    stack.append(step.step.id)
                    continue

                if step.step.op is server_lib.Opcode.PLANNED:
                    planned.add(step.step.id)
                elif step.step.op is server_lib.Opcode.SLEEP:
                    steps[step.step.id] = None
                    stack.append(step.step.id)
                elif step.step.op is server_lib.Opcode.STEP_RUN:
                    steps[step.step.id] = step.output
                    stack.append(step.step.id)
                else:
                    raise UnstubbedStepError(step.step.display_name)

            continue

        return _Result(
            error=None,
            output=res.output,
            status=Status.COMPLETED,
        )


@dataclasses.dataclass
class _Result:
    error: Exception | None
    output: object
    status: Status



import typing
from urllib.parse import urljoin

from inngest._internal import client_lib, types


class _TokenResponse(types.BaseModel):
    jwt: str


async def get_subscription_token(
    client: client_lib.Inngest,
    channel: str,
    topics: list[str],
) -> typing.Mapping[str, object]:
    """
    Create a subscription token for a given channel and topics.
    The token can be used by a client to subscribe to realtime events,
    including front-end applications using the @inngest/realtime npm package.
    """
    data = []
    for topic in topics:
        data.append(
            {
                "channel": channel,
                "name": topic,
                "kind": "run",
            }
        )

    res = await client._http_client.post(
        url=urljoin(client._api_origin, "/v1/realtime/token"),
        body=data,
    )
    if isinstance(res, Exception):
        raise res
    # Response is an object with a "jwt" property which is a string
    response_data = _TokenResponse.model_validate(res.json())

    # Return a dictionary ready to be used by the @inngest/realtime npm package
    return {
        "channel": channel,
        "topics": topics,
        "key": response_data.jwt,
    }


def get_subscription_token_sync(
    client: client_lib.Inngest,
    channel: str,
    topics: list[str],
) -> typing.Mapping[str, object]:
    """
    Create a subscription token for a given channel and topics synchronously.
    The token can be used by a client to subscribe to realtime events,
    including front-end applications using the @inngest/realtime npm package.
    """
    data = []
    for topic in topics:
        data.append(
            {
                "channel": channel,
                "name": topic,
                "kind": "run",
            }
        )

    res = client._http_client.post_sync(
        url=urljoin(client._api_origin, "/v1/realtime/token"),
        body=data,
    )
    if isinstance(res, Exception):
        raise res
    # Response is an object with a "jwt" property which is a string
    response_data = _TokenResponse.model_validate(res.json())

    # Return a dictionary ready to be used by the @inngest/realtime npm package
    return {
        "channel": channel,
        "topics": topics,
        "key": response_data.jwt,
    }



from __future__ import annotations

import typing

import inngest
from inngest._internal import server_lib


class StateDriver(typing.Protocol):
    """
    Protocol for the state driver.
    """

    def load_steps(self, steps: inngest.StepMemos) -> None:
        """
        Retrieve the value associated with the key.

        Args:
        ----
            steps: Steps whose output may need to be loaded from the remote store.
        """

        ...

    def save_step(
        self,
        run_id: str,
        value: object,
    ) -> dict[str, object]:
        """
        Store the value and return a key to retrieve it later.

        Args:
        ----
            run_id: Run ID.
            value: Output for an ended step.
        """

        ...


class RemoteStateMiddleware(inngest.MiddlewareSync):
    """
    Middleware that reads/writes step output in a custom store (e.g. AWS S3).
    This can drastically reduce bandwidth to/from the Inngest server, since step
    output is stored within your infrastructure rather than Inngest's.
    """

    _run_id: str | None = None

    def __init__(
        self,
        client: inngest.Inngest,
        raw_request: object,
        driver: StateDriver,
    ) -> None:
        """
        Args:
        ----
            client: Inngest client.
            raw_request: Framework/platform specific request object.
            driver: State driver.
        """

        super().__init__(client, raw_request)

        self._driver = driver

    @classmethod
    def factory(
        cls,
        driver: StateDriver,
    ) -> typing.Callable[[inngest.Inngest, object], RemoteStateMiddleware]:
        """
        Create a remote state middleware that can be passed to an Inngest client
        or function.

        Args:
        ----
            driver: State driver.
        """

        def _factory(
            client: inngest.Inngest,
            raw_request: object,
        ) -> RemoteStateMiddleware:
            return cls(
                client,
                raw_request,
                driver,
            )

        return _factory

    def transform_input(
        self,
        ctx: inngest.Context | inngest.ContextSync,
        function: inngest.Function[typing.Any],
        steps: inngest.StepMemos,
    ) -> None:
        """
        Inject remote state.
        """

        self._driver.load_steps(steps)
        self._run_id = ctx.run_id

    def transform_output(self, result: inngest.TransformOutputResult) -> None:
        """
        Store step output externally and replace with a marker and key.
        """

        if result.step is None:
            return None

        if result.step.op is not server_lib.Opcode.STEP_RUN:
            return None

        if result.has_output() is False:
            return None

        if self._run_id is None:
            # Unreachable
            raise Exception("missing run ID")

        result.output = self._driver.save_step(
            self._run_id,
            result.output,
        )



from .handler import CommHandler, get_function_configs
from .models import CommRequest, CommResponse

__all__ = [
    "CommHandler",
    "CommRequest",
    "CommResponse",
    "get_function_configs",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: connect.proto
# Protobuf Python Version: 5.29.4
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    4,
    '',
    'connect.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rconnect.proto\x12\nconnect.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"O\n\x0e\x43onnectMessage\x12,\n\x04kind\x18\x01 \x01(\x0e\x32\x1e.connect.v1.GatewayMessageType\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"a\n\x10\x41ppConfiguration\x12\x10\n\x08\x61pp_name\x18\x01 \x01(\t\x12\x18\n\x0b\x61pp_version\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x11\n\tfunctions\x18\x04 \x01(\x0c\x42\x0e\n\x0c_app_version\"5\n\x08\x41uthData\x12\x15\n\rsession_token\x18\x01 \x01(\t\x12\x12\n\nsync_token\x18\x02 \x01(\t\"\x8b\x04\n\x18WorkerConnectRequestData\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x13\n\x0binstance_id\x18\x02 \x01(\t\x12\'\n\tauth_data\x18\x03 \x01(\x0b\x32\x14.connect.v1.AuthData\x12\x14\n\x0c\x63\x61pabilities\x18\x04 \x01(\x0c\x12*\n\x04\x61pps\x18\x05 \x03(\x0b\x32\x1c.connect.v1.AppConfiguration\x12#\n\x1bworker_manual_readiness_ack\x18\x06 \x01(\x08\x12\x37\n\x11system_attributes\x18\x07 \x01(\x0b\x32\x1c.connect.v1.SystemAttributes\x12\x18\n\x0b\x65nvironment\x18\x08 \x01(\tH\x00\x88\x01\x01\x12\x11\n\tframework\x18\t \x01(\t\x12\x15\n\x08platform\x18\n \x01(\tH\x01\x88\x01\x01\x12\x13\n\x0bsdk_version\x18\x0b \x01(\t\x12\x14\n\x0csdk_language\x18\x0c \x01(\t\x12.\n\nstarted_at\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12#\n\x16max_worker_concurrency\x18\x0e \x01(\x03H\x02\x88\x01\x01\x42\x0e\n\x0c_environmentB\x0b\n\t_platformB\x19\n\x17_max_worker_concurrency\"W\n\x1aGatewayConnectionReadyData\x12\x1a\n\x12heartbeat_interval\x18\x01 \x01(\t\x12\x1d\n\x15\x65xtend_lease_interval\x18\x02 \x01(\t\"\xb1\x02\n\x1aGatewayExecutorRequestData\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12\x0e\n\x06\x65nv_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12\x10\n\x08\x61pp_name\x18\x05 \x01(\t\x12\x13\n\x0b\x66unction_id\x18\x06 \x01(\t\x12\x15\n\rfunction_slug\x18\x07 \x01(\t\x12\x14\n\x07step_id\x18\x08 \x01(\tH\x00\x88\x01\x01\x12\x17\n\x0frequest_payload\x18\t \x01(\x0c\x12\x18\n\x10system_trace_ctx\x18\n \x01(\x0c\x12\x16\n\x0euser_trace_ctx\x18\x0b \x01(\x0c\x12\x0e\n\x06run_id\x18\x0c \x01(\t\x12\x10\n\x08lease_id\x18\r \x01(\tB\n\n\x08_step_id\"\xd9\x01\n\x14WorkerRequestAckData\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12\x0e\n\x06\x65nv_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12\x15\n\rfunction_slug\x18\x05 \x01(\t\x12\x14\n\x07step_id\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x10system_trace_ctx\x18\x07 \x01(\x0c\x12\x16\n\x0euser_trace_ctx\x18\x08 \x01(\x0c\x12\x0e\n\x06run_id\x18\t \x01(\tB\n\n\x08_step_id\"\xf3\x01\n\x1cWorkerRequestExtendLeaseData\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12\x0e\n\x06\x65nv_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12\x15\n\rfunction_slug\x18\x05 \x01(\t\x12\x14\n\x07step_id\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x10system_trace_ctx\x18\x07 \x01(\x0c\x12\x16\n\x0euser_trace_ctx\x18\x08 \x01(\x0c\x12\x0e\n\x06run_id\x18\t \x01(\t\x12\x10\n\x08lease_id\x18\n \x01(\tB\n\n\x08_step_id\"\xac\x01\n\x1fWorkerRequestExtendLeaseAckData\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12\x0e\n\x06\x65nv_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12\x15\n\rfunction_slug\x18\x05 \x01(\t\x12\x19\n\x0cnew_lease_id\x18\x06 \x01(\tH\x00\x88\x01\x01\x42\x0f\n\r_new_lease_id\"\xbe\x02\n\x0bSDKResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x12\n\naccount_id\x18\x02 \x01(\t\x12\x0e\n\x06\x65nv_id\x18\x03 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x04 \x01(\t\x12-\n\x06status\x18\x05 \x01(\x0e\x32\x1d.connect.v1.SDKResponseStatus\x12\x0c\n\x04\x62ody\x18\x06 \x01(\x0c\x12\x10\n\x08no_retry\x18\x07 \x01(\x08\x12\x18\n\x0bretry_after\x18\x08 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x0bsdk_version\x18\t \x01(\t\x12\x17\n\x0frequest_version\x18\n \x01(\r\x12\x18\n\x10system_trace_ctx\x18\x0b \x01(\x0c\x12\x16\n\x0euser_trace_ctx\x18\x0c \x01(\x0c\x12\x0e\n\x06run_id\x18\r \x01(\tB\x0e\n\x0c_retry_after\"(\n\x12WorkerReplyAckData\x12\x12\n\nrequest_id\x18\x01 \x01(\t\"\x92\x04\n\x0c\x43onnMetadata\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ngateway_id\x18\x02 \x01(\t\x12\x13\n\x0binstance_id\x18\x03 \x01(\t\x12H\n\x11\x61ll_worker_groups\x18\x04 \x03(\x0b\x32-.connect.v1.ConnMetadata.AllWorkerGroupsEntry\x12N\n\x14synced_worker_groups\x18\x05 \x03(\x0b\x32\x30.connect.v1.ConnMetadata.SyncedWorkerGroupsEntry\x12,\n\x06status\x18\x06 \x01(\x0e\x32\x1c.connect.v1.ConnectionStatus\x12\x35\n\x11last_heartbeat_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0csdk_language\x18\x08 \x01(\t\x12\x13\n\x0bsdk_version\x18\t \x01(\t\x12\x30\n\nattributes\x18\n \x01(\x0b\x32\x1c.connect.v1.SystemAttributes\x1a\x36\n\x14\x41llWorkerGroupsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x39\n\x17SyncedWorkerGroupsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"D\n\x10SystemAttributes\x12\x11\n\tcpu_cores\x18\x01 \x01(\x05\x12\x11\n\tmem_bytes\x18\x02 \x01(\x03\x12\n\n\x02os\x18\x03 \x01(\t\"\xc0\x01\n\tConnGroup\x12\x0e\n\x06\x65nv_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61pp_id\x18\x02 \x01(\t\x12\x10\n\x08\x61pp_name\x18\x03 \x01(\t\x12\x0c\n\x04hash\x18\x04 \x01(\t\x12\'\n\x05\x63onns\x18\x05 \x03(\x0b\x32\x18.connect.v1.ConnMetadata\x12\x14\n\x07sync_id\x18\x06 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0b\x61pp_version\x18\x07 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_sync_idB\x0e\n\x0c_app_version\"\x82\x01\n\rStartResponse\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\x18\n\x10gateway_endpoint\x18\x02 \x01(\t\x12\x15\n\rgateway_group\x18\x03 \x01(\t\x12\x15\n\rsession_token\x18\x04 \x01(\t\x12\x12\n\nsync_token\x18\x05 \x01(\t\"(\n\x0cStartRequest\x12\x18\n\x10\x65xclude_gateways\x18\x01 \x03(\t\"#\n\rFlushResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\"\x99\x01\n\x10PubSubAckMessage\x12&\n\x02ts\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x11\n\x04nack\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x31\n\x0bnack_reason\x18\x03 \x01(\x0b\x32\x17.connect.v1.SystemErrorH\x01\x88\x01\x01\x42\x07\n\x05_nackB\x0e\n\x0c_nack_reason\"H\n\x0bSystemError\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x11\n\x04\x64\x61ta\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x12\x0f\n\x07message\x18\x03 \x01(\tB\x07\n\x05_data*\xf4\x02\n\x12GatewayMessageType\x12\x11\n\rGATEWAY_HELLO\x10\x00\x12\x12\n\x0eWORKER_CONNECT\x10\x01\x12\x0f\n\x0bSYNC_FAILED\x10\x0e\x12\x1c\n\x18GATEWAY_CONNECTION_READY\x10\x02\x12\x1c\n\x18GATEWAY_EXECUTOR_REQUEST\x10\x03\x12\x10\n\x0cWORKER_READY\x10\x04\x12\x16\n\x12WORKER_REQUEST_ACK\x10\x05\x12\x1f\n\x1bWORKER_REQUEST_EXTEND_LEASE\x10\x0c\x12#\n\x1fWORKER_REQUEST_EXTEND_LEASE_ACK\x10\r\x12\x10\n\x0cWORKER_REPLY\x10\x06\x12\x14\n\x10WORKER_REPLY_ACK\x10\x07\x12\x10\n\x0cWORKER_PAUSE\x10\x08\x12\x14\n\x10WORKER_HEARTBEAT\x10\t\x12\x15\n\x11GATEWAY_HEARTBEAT\x10\n\x12\x13\n\x0fGATEWAY_CLOSING\x10\x0b*;\n\x11SDKResponseStatus\x12\x11\n\rNOT_COMPLETED\x10\x00\x12\x08\n\x04\x44ONE\x10\x01\x12\t\n\x05\x45RROR\x10\x02*_\n\x10\x43onnectionStatus\x12\r\n\tCONNECTED\x10\x00\x12\t\n\x05READY\x10\x01\x12\x0c\n\x08\x44RAINING\x10\x02\x12\x11\n\rDISCONNECTING\x10\x03\x12\x10\n\x0c\x44ISCONNECTED\x10\x04*\x8d\x01\n\x16WorkerDisconnectReason\x12\x13\n\x0fWORKER_SHUTDOWN\x10\x00\x12\x0e\n\nUNEXPECTED\x10\x01\x12\x14\n\x10GATEWAY_DRAINING\x10\x02\x12!\n\x1d\x43ONSECUTIVE_HEARTBEATS_MISSED\x10\x03\x12\x15\n\x11MESSAGE_TOO_LARGE\x10\x04\x42\x39Z7github.com/inngest/inngest/proto/gen/connect/v1;connectb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'connect_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z7github.com/inngest/inngest/proto/gen/connect/v1;connect'
  _globals['_CONNMETADATA_ALLWORKERGROUPSENTRY']._loaded_options = None
  _globals['_CONNMETADATA_ALLWORKERGROUPSENTRY']._serialized_options = b'8\001'
  _globals['_CONNMETADATA_SYNCEDWORKERGROUPSENTRY']._loaded_options = None
  _globals['_CONNMETADATA_SYNCEDWORKERGROUPSENTRY']._serialized_options = b'8\001'
  _globals['_GATEWAYMESSAGETYPE']._serialized_start=3465
  _globals['_GATEWAYMESSAGETYPE']._serialized_end=3837
  _globals['_SDKRESPONSESTATUS']._serialized_start=3839
  _globals['_SDKRESPONSESTATUS']._serialized_end=3898
  _globals['_CONNECTIONSTATUS']._serialized_start=3900
  _globals['_CONNECTIONSTATUS']._serialized_end=3995
  _globals['_WORKERDISCONNECTREASON']._serialized_start=3998
  _globals['_WORKERDISCONNECTREASON']._serialized_end=4139
  _globals['_CONNECTMESSAGE']._serialized_start=62
  _globals['_CONNECTMESSAGE']._serialized_end=141
  _globals['_APPCONFIGURATION']._serialized_start=143
  _globals['_APPCONFIGURATION']._serialized_end=240
  _globals['_AUTHDATA']._serialized_start=242
  _globals['_AUTHDATA']._serialized_end=295
  _globals['_WORKERCONNECTREQUESTDATA']._serialized_start=298
  _globals['_WORKERCONNECTREQUESTDATA']._serialized_end=821
  _globals['_GATEWAYCONNECTIONREADYDATA']._serialized_start=823
  _globals['_GATEWAYCONNECTIONREADYDATA']._serialized_end=910
  _globals['_GATEWAYEXECUTORREQUESTDATA']._serialized_start=913
  _globals['_GATEWAYEXECUTORREQUESTDATA']._serialized_end=1218
  _globals['_WORKERREQUESTACKDATA']._serialized_start=1221
  _globals['_WORKERREQUESTACKDATA']._serialized_end=1438
  _globals['_WORKERREQUESTEXTENDLEASEDATA']._serialized_start=1441
  _globals['_WORKERREQUESTEXTENDLEASEDATA']._serialized_end=1684
  _globals['_WORKERREQUESTEXTENDLEASEACKDATA']._serialized_start=1687
  _globals['_WORKERREQUESTEXTENDLEASEACKDATA']._serialized_end=1859
  _globals['_SDKRESPONSE']._serialized_start=1862
  _globals['_SDKRESPONSE']._serialized_end=2180
  _globals['_WORKERREPLYACKDATA']._serialized_start=2182
  _globals['_WORKERREPLYACKDATA']._serialized_end=2222
  _globals['_CONNMETADATA']._serialized_start=2225
  _globals['_CONNMETADATA']._serialized_end=2755
  _globals['_CONNMETADATA_ALLWORKERGROUPSENTRY']._serialized_start=2642
  _globals['_CONNMETADATA_ALLWORKERGROUPSENTRY']._serialized_end=2696
  _globals['_CONNMETADATA_SYNCEDWORKERGROUPSENTRY']._serialized_start=2698
  _globals['_CONNMETADATA_SYNCEDWORKERGROUPSENTRY']._serialized_end=2755
  _globals['_SYSTEMATTRIBUTES']._serialized_start=2757
  _globals['_SYSTEMATTRIBUTES']._serialized_end=2825
  _globals['_CONNGROUP']._serialized_start=2828
  _globals['_CONNGROUP']._serialized_end=3020
  _globals['_STARTRESPONSE']._serialized_start=3023
  _globals['_STARTRESPONSE']._serialized_end=3153
  _globals['_STARTREQUEST']._serialized_start=3155
  _globals['_STARTREQUEST']._serialized_end=3195
  _globals['_FLUSHRESPONSE']._serialized_start=3197
  _globals['_FLUSHRESPONSE']._serialized_end=3232
  _globals['_PUBSUBACKMESSAGE']._serialized_start=3235
  _globals['_PUBSUBACKMESSAGE']._serialized_end=3388
  _globals['_SYSTEMERROR']._serialized_start=3390
  _globals['_SYSTEMERROR']._serialized_end=3462
# @@protoc_insertion_point(module_scope)



import unittest.mock

import pytest

from . import get_step_context, set_step_context, step


class TestContextVar:
    def test_set_and_get(self) -> None:
        step = unittest.mock.Mock()
        with set_step_context(step):
            assert get_step_context() == step

    def test_get_outside_with(self) -> None:
        step = unittest.mock.Mock()
        with set_step_context(step):
            pass

        with pytest.raises(LookupError):
            get_step_context()

    def test_set_without_with(self) -> None:
        step = unittest.mock.Mock()
        set_step_context(step)

        with pytest.raises(LookupError):
            get_step_context()

    def test_get_without_set(self) -> None:
        with pytest.raises(LookupError):
            get_step_context()

    def test_nested_set(self) -> None:
        with set_step_context(unittest.mock.Mock()):
            with pytest.raises(Exception) as e:
                with set_step_context(unittest.mock.Mock()):
                    pass
            assert str(e.value) == "Step context already set"


class TestStepDecorator:
    def test_can_call_outside_of_step_context(self) -> None:
        """
        If the step context is not set, the decorator is effectively a no-op (it
        just calls the decorated function directly)
        """

        @step("my-step")
        def my_step(name: str) -> str:
            return f"Hello {name}"

        assert my_step("Alice") == "Hello Alice"



import datetime

from .registration import (
    Batch,
    Cancel,
    Concurrency,
    Debounce,
    FunctionConfig,
    Priority,
    RateLimit,
    Retries,
    Runtime,
    Singleton,
    Step,
    Throttle,
    Timeouts,
    TriggerCron,
    TriggerEvent,
)


def test_serialization() -> None:
    data = FunctionConfig(
        batch_events=Batch(
            max_size=10,
            timeout=datetime.timedelta(seconds=60),
            key="foo",
            if_exp="bar",
        ),
        cancel=[
            Cancel(
                event="foo",
                if_exp="foo",
                timeout=datetime.timedelta(seconds=60),
            )
        ],
        concurrency=[
            Concurrency(
                key="foo",
                limit=1,
                scope="account",
            )
        ],
        debounce=Debounce(
            key="foo",
            period=datetime.timedelta(seconds=60),
            timeout=datetime.timedelta(minutes=2),
        ),
        id="foo",
        idempotency="event.data.foo",
        name="foo",
        priority=Priority(
            run="event.data.plan == 'enterprise' ? 180 : 0",
        ),
        steps={
            "foo": Step(
                id="foo",
                name="foo",
                retries=Retries(attempts=1),
                runtime=Runtime(type="http", url="foo"),
            )
        },
        rate_limit=RateLimit(
            key="foo",
            limit=1,
            period=datetime.timedelta(seconds=60),
        ),
        throttle=Throttle(
            key="foo",
            limit=1,
            period=datetime.timedelta(seconds=60),
        ),
        timeouts=Timeouts(
            start=datetime.timedelta(minutes=5),
            finish=datetime.timedelta(hours=1),
        ),
        singleton=Singleton(
            key="foo",
            mode="skip",
        ),
        triggers=[
            TriggerCron(cron="foo"),
            TriggerEvent(event="foo", expression="foo"),
        ],
    ).to_dict()
    if isinstance(data, Exception):
        raise data

    assert data == {
        "batchEvents": {
            "maxSize": 10,
            "timeout": "1m",
            "key": "foo",
            "if": "bar",
        },
        "cancel": [
            {
                "event": "foo",
                "if": "foo",
                "timeout": "1m",
            }
        ],
        "concurrency": [
            {
                "key": "foo",
                "limit": 1,
                "scope": "account",
            }
        ],
        "debounce": {
            "key": "foo",
            "period": "1m",
            "timeout": "2m",
        },
        "id": "foo",
        "idempotency": "event.data.foo",
        "name": "foo",
        "priority": {
            "run": "event.data.plan == 'enterprise' ? 180 : 0",
        },
        "rateLimit": {
            "key": "foo",
            "limit": 1,
            "period": "1m",
        },
        "steps": {
            "foo": {
                "id": "foo",
                "name": "foo",
                "retries": {"attempts": 1},
                "runtime": {
                    "type": "http",
                    "url": "foo",
                },
            }
        },
        "throttle": {
            "key": "foo",
            "limit": 1,
            "burst": 1,
            "period": "1m",
        },
        "timeouts": {
            "start": "5m",
            "finish": "1h",
        },
        "singleton": {
            "key": "foo",
            "mode": "skip",
        },
        "triggers": [
            {"cron": "foo"},
            {
                "event": "foo",
                "expression": "foo",
            },
        ],
    }



"""
Connect creates a persistent outbound connection to Inngest. Compared to our "serve" functions, Connect:
- Doesn't require exposing your app to the public internet.
- Is lower latency.
- Avoids HTTP timeouts.

Note that Connect does not work with serverless platforms: you must run your app
in a long-lived process. If you're using a serverless platform, you should use
our "serve" functions instead.
"""

from ._internal.connect import connect
from ._internal.models import ConnectionState

__all__ = ["connect", "ConnectionState"]



from __future__ import annotations

import datetime
import typing

import typing_extensions

from inngest._internal import errors, server_lib, transforms, types
from inngest._internal.client_lib import models as client_models

from . import base

# Avoid circular import at runtime
if typing.TYPE_CHECKING:
    from inngest._internal import (
        client_lib,
        execution_lib,
        function,
        middleware_lib,
    )
    from inngest.experimental import ai


class StepSync(base.StepBase):
    def __init__(
        self,
        client: client_lib.Inngest,
        exe: execution_lib.BaseExecutionSync,
        middleware: middleware_lib.MiddlewareManager,
        step_id_counter: base.StepIDCounter,
        target_hashed_id: str | None,
    ) -> None:
        super().__init__(
            client,
            middleware,
            step_id_counter,
            target_hashed_id,
        )

        self.ai = AI(self)
        self._execution = exe

    def invoke(
        self,
        step_id: str,
        *,
        function: function.Function[types.T],
        data: typing.Mapping[str, object] | None = None,
        timeout: int | datetime.timedelta | None = None,
        v: str | None = None,
    ) -> types.T:
        """
        Invoke an Inngest function with data. Returns the result of the returned
        value of the function or `None` if the function does not return a value.

        If a function isn't found or otherwise errors, the step will fail and
        raise a `NonRetriableError`.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.

            function: The function object to invoke.
            data: Will become `event.data` in the invoked function. Must be JSON serializable.
            timeout: The maximum number of milliseconds to wait for the function to complete.
            v: Will become `event.v` in the invoked function.
        """

        output = self.invoke_by_id(
            step_id,
            app_id=self._client.app_id,
            function_id=function._opts.local_id,
            data=data,
            timeout=timeout,
            v=v,
        )

        output = self._client._deserialize(output, function._output_type)

        return output  # type: ignore[return-value]

    def invoke_by_id(
        self,
        step_id: str,
        *,
        app_id: str | None = None,
        function_id: str,
        data: typing.Mapping[str, object] | None = None,
        timeout: int | datetime.timedelta | None = None,
        v: str | None = None,
    ) -> object:
        """
        Invoke an Inngest function with data. Returns the result of the returned
        value of the function or `None` if the function does not return a value.

        If app ID is not specified, the invoked function must be in the same
        app.

        If a function isn't found or otherwise errors, the step will fail and
        raise a `NonRetriableError`.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.

            app_id: The app ID of the function to invoke. Not necessary if this function and the invoked function are in the same app.
            function_id: The ID of the function to invoke.
            data: Will become `event.data` in the invoked function. Must be JSON serializable.
            timeout: The maximum number of milliseconds to wait for the function to complete.
            v: Will become `event.v` in the invoked function.
        """

        if app_id is None:
            app_id = self._client.app_id

        parsed_step_id = self._parse_step_id(step_id)

        timeout_str = transforms.to_maybe_duration_str(timeout)
        if isinstance(timeout_str, Exception):
            raise timeout_str

        opts = base.InvokeOpts(
            function_id=f"{app_id}-{function_id}",
            payload=base.InvokeOptsPayload(
                data=data,
                v=v,
            ),
            timeout=timeout_str,
        ).to_dict()
        if isinstance(opts, Exception):
            raise opts

        step_info = base.StepInfo(
            display_name=parsed_step_id.user_facing,
            id=parsed_step_id.hashed,
            name=parsed_step_id.user_facing,
            op=server_lib.Opcode.INVOKE,
            opts=opts,
            userland=parsed_step_id.userland_info(),
        )

        with self._execution.report_step(step_info) as step:
            if step.skip:
                raise base.SkipInterrupt(parsed_step_id.user_facing)
            if step.error is not None:
                raise step.error
            elif not isinstance(step.output, types.EmptySentinel):
                return step.output

        raise Exception("unreachable")

    def run(
        self,
        step_id: str,
        handler: typing.Callable[
            [typing_extensions.Unpack[types.TTuple]],
            types.T,
        ],
        *handler_args: typing_extensions.Unpack[types.TTuple],
        output_type: object = types.EmptySentinel,
    ) -> types.T:
        """
        Run logic that should be retried on error and memoized after success.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            handler: The logic to run.
            *handler_args: Arguments to pass to the handler.
            output_type: Only set if returning a non-JSON-serializable object. Related to the client's serializer argument.
        """

        parsed_step_id = self._parse_step_id(step_id)

        step_info = base.StepInfo(
            display_name=parsed_step_id.user_facing,
            id=parsed_step_id.hashed,
            name=parsed_step_id.user_facing,
            op=server_lib.Opcode.STEP_RUN,
            userland=parsed_step_id.userland_info(),
        )

        with self._execution.report_step(step_info) as step:
            if step.skip:
                raise base.SkipInterrupt(parsed_step_id.user_facing)
            if step.error is not None:
                raise step.error
            elif not isinstance(step.output, types.EmptySentinel):
                return self._client._deserialize(step.output, output_type)  # type: ignore[return-value]

            try:
                output = handler(*handler_args)
                output = self._client._serialize(output, output_type)  # type: ignore[assignment]

                raise base.ResponseInterrupt(
                    base.StepResponse(
                        output=output,
                        step=step_info,
                    )
                )
            except (errors.NonRetriableError, errors.RetryAfterError) as err:
                # Bubble up these error types to the function level
                raise err
            except Exception as err:
                transforms.remove_first_traceback_frame(err)

                max_attempts = self._execution._request.ctx.max_attempts
                if (
                    max_attempts is not None
                    and self._execution._request.ctx.attempt + 1 >= max_attempts
                ):
                    step_info.op = server_lib.Opcode.STEP_FAILED
                else:
                    step_info.op = server_lib.Opcode.STEP_ERROR

                raise base.ResponseInterrupt(
                    base.StepResponse(
                        original_error=err,
                        step=step_info,
                    )
                )
            except base.NestedStepInterrupt:
                step_info.op = server_lib.Opcode.STEP_ERROR
                raise base.ResponseInterrupt(
                    base.StepResponse(
                        original_error=errors.CodedError(
                            server_lib.ErrorCode.STEP_NESTED,
                            "Nested steps are not supported.",
                            is_retriable=False,
                        ),
                        step=step_info,
                    )
                )

    def send_event(
        self,
        step_id: str,
        events: server_lib.Event | list[server_lib.Event],
    ) -> list[str]:
        """
        Send an event or list of events.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            events: An event or list of events to send.
        """

        def fn() -> list[str]:
            if isinstance(events, list):
                _events = events
            else:
                _events = [events]

            middleware_err = self._middleware.before_send_events_sync(_events)
            if isinstance(middleware_err, Exception):
                raise middleware_err

            # Need to initialize `result` because of the `finally` block.
            result: client_models.SendEventsResult | None = None

            try:
                result = client_models.SendEventsResult(
                    ids=self._client.send_sync(
                        events,
                        # Skip middleware since we're already running it above. Without
                        # this, we'll double-call middleware hooks
                        skip_middleware=True,
                    )
                )
            except errors.SendEventsError as err:
                result = client_models.SendEventsResult(
                    error=str(err),
                    ids=err.ids,
                )
                raise err
            except Exception as err:
                result = client_models.SendEventsResult(
                    error=str(err),
                    ids=[],
                )
                raise err
            finally:
                if result is not None:
                    middleware_err = self._middleware.after_send_events_sync(
                        result
                    )
                    if isinstance(middleware_err, Exception):
                        raise middleware_err

            if result is None:
                raise Exception("unreachable")

            return result.ids

        return self.run(step_id, fn)

    def sleep(
        self,
        step_id: str,
        duration: int | datetime.timedelta,
    ) -> None:
        """
        Sleep for a duration.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            duration: The number of milliseconds to sleep.
        """
        if isinstance(duration, int):
            until = datetime.datetime.now() + datetime.timedelta(
                milliseconds=duration
            )
        else:
            until = datetime.datetime.now() + duration

        return self.sleep_until(step_id, until)

    def sleep_until(
        self,
        step_id: str,
        until: datetime.datetime,
    ) -> None:
        """
        Sleep until a specific time.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            until: The time to sleep until.
        """

        parsed_step_id = self._parse_step_id(step_id)

        step_info = base.StepInfo(
            display_name=parsed_step_id.user_facing,
            id=parsed_step_id.hashed,
            name=transforms.to_iso_utc(until),
            op=server_lib.Opcode.SLEEP,
            userland=parsed_step_id.userland_info(),
        )

        with self._execution.report_step(step_info) as step:
            if step.skip:
                raise base.SkipInterrupt(parsed_step_id.user_facing)
            if step.error is not None:
                raise step.error
            elif not isinstance(step.output, types.EmptySentinel):
                return step.output  # type: ignore

        raise Exception("unreachable")

    def wait_for_event(
        self,
        step_id: str,
        *,
        event: str,
        if_exp: str | None = None,
        timeout: int | datetime.timedelta,
    ) -> server_lib.Event | None:
        """
        Wait for an event to be sent.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            event: Event name.
            if_exp: An expression to filter events.
            timeout: The maximum number of milliseconds to wait for the event.
        """

        parsed_step_id = self._parse_step_id(step_id)

        timeout_str = transforms.to_duration_str(timeout)
        if isinstance(timeout_str, Exception):
            raise timeout_str

        opts = base.WaitForEventOpts(
            if_exp=if_exp,
            timeout=timeout_str,
        ).to_dict()
        if isinstance(opts, Exception):
            raise opts

        step_info = base.StepInfo(
            id=parsed_step_id.hashed,
            display_name=parsed_step_id.user_facing,
            name=event,
            op=server_lib.Opcode.WAIT_FOR_EVENT,
            opts=opts,
            userland=parsed_step_id.userland_info(),
        )

        with self._execution.report_step(step_info) as step:
            if step.skip:
                raise base.SkipInterrupt(parsed_step_id.user_facing)
            if step.error is not None:
                raise step.error
            elif not isinstance(step.output, types.EmptySentinel):
                if step.output is None:
                    # Timeout
                    return None

                # Fulfilled by an event
                return server_lib.Event.model_validate(step.output)

        raise Exception("unreachable")


class AI:
    def __init__(self, step: StepSync) -> None:
        self._step = step

    def infer(
        self,
        step_id: str,
        *,
        adapter: ai.BaseAdapter,
        body: dict[str, typing.Any],
    ) -> dict[str, object]:
        """
        EXPERIMENTAL: This method's interface is subject to change. Make an AI
        model call using Inngest as an intermediary. The request-sending
        responsibility is offloaded to the Inngest server, so you aren't using
        compute while waiting for a response.

        Args:
        ----
            step_id: Durable step ID. Should usually be unique within a function, but it's OK to reuse as long as your function is deterministic.
            adapter: AI model adapter.
            body: Request body sent to the AI model.
        """

        parsed_step_id = self._step._parse_step_id(step_id)

        adapter.on_call(body)

        opts = base.AIInferOpts(
            auth_key=adapter.auth_key(),
            body=body,
            format=adapter.format(),
            headers=adapter.headers(),
            url=adapter.url_infer(),
        ).to_dict()
        if isinstance(opts, Exception):
            raise opts

        step_info = base.StepInfo(
            display_name=parsed_step_id.user_facing,
            id=parsed_step_id.hashed,
            name=parsed_step_id.user_facing,
            op=server_lib.Opcode.AI_GATEWAY,
            opts=opts,
            userland=parsed_step_id.userland_info(),
        )

        with self._step._execution.report_step(step_info) as step:
            if step.skip:
                raise base.SkipInterrupt(parsed_step_id.user_facing)
            if step.error is not None:
                raise step.error
            elif not isinstance(step.output, types.EmptySentinel):
                return step.output  # type: ignore[return-value]

        raise Exception("unreachable")



from __future__ import annotations

import asyncio
import time
import unittest

import pytest

from .async_lib import run_sync


class TestRunSync(unittest.IsolatedAsyncioTestCase):
    async def test_with_running_loop(self) -> None:
        sleep_dur = 0.5
        result_holder: list[str] = []

        async def fn() -> str:
            await asyncio.sleep(sleep_dur)
            result_holder.append("completed")
            return "output"

        # Immediately returns
        with AssertDuration(0):
            future = run_sync(fn())

        assert not isinstance(future, Exception)

        # Returned immediately without blocking
        assert len(result_holder) == 0

        # Wait for the task to complete
        with AssertDuration(sleep_dur):
            output = await asyncio.wrap_future(future)
            assert output == "output"

        assert result_holder == ["completed"]

    async def test_with_exception(self) -> None:
        sleep_dur = 0.5

        async def fn() -> None:
            await asyncio.sleep(sleep_dur)
            raise ValueError("test error")

        future = run_sync(fn())
        assert not isinstance(future, Exception)

        with AssertDuration(sleep_dur):
            with pytest.raises(ValueError, match="test error"):
                await asyncio.wrap_future(future)

    async def test_multiple_tasks(self) -> None:
        sleep_dur = 0.5
        results: list[int] = []

        async def fn(value: int) -> None:
            await asyncio.sleep(sleep_dur)
            results.append(value)

        # Schedule multiple tasks
        future_1 = run_sync(fn(1))
        future_2 = run_sync(fn(2))
        future_3 = run_sync(fn(3))
        assert not isinstance(future_1, Exception)
        assert not isinstance(future_2, Exception)
        assert not isinstance(future_3, Exception)
        assert len(results) == 0

        with AssertDuration(sleep_dur):
            await asyncio.gather(
                asyncio.wrap_future(future_1),
                asyncio.wrap_future(future_2),
                asyncio.wrap_future(future_3),
            )

        # All tasks should have completed
        assert sorted(results) == [1, 2, 3]

    def test_no_event_loop(self) -> None:
        """
        Errors when no event loop is running. The missing event loop scenario is
        implicitly set up because the test method is not async
        """

        async def fn() -> None:
            pass

        result = run_sync(fn())

        assert isinstance(result, RuntimeError)


class AssertDuration:
    """
    Helper for asserting that a block of code takes a certain amount of time.
    This is a context manager, so use it in a `with` statement
    """

    def __init__(
        self,
        expected: float,
        *,
        tolerance: float = 0.2,
    ) -> None:
        """
        Args:
        ----
            expected: The expected duration in seconds.
            tolerance: The tolerance in seconds. This is only applied to the maximum duration.
        """

        self._expected = expected
        self._start_time: float = 0
        self._tolerance = tolerance

    def __enter__(self) -> AssertDuration:
        self._start_time = time.time()
        return self

    def __exit__(self, *args: object) -> None:
        elapsed = time.time() - self._start_time

        assert elapsed >= self._expected
        assert elapsed <= self._expected + self._tolerance


class TestAssertDuration:
    def test_assert_duration(self) -> None:
        with AssertDuration(0.1):
            time.sleep(0.1)

        with AssertDuration(0.1):
            time.sleep(0.2)

        with pytest.raises(AssertionError):
            with AssertDuration(0.1):
                time.sleep(0.05)



import asyncio
import enum
import platform
import re

import psutil
import pydantic_core

from inngest._internal import const, errors, server_lib, types

from . import async_lib, connect_pb2, pb_utils, ws_utils
from .base_handler import BaseHandler
from .models import AppConfig, ConnectionState, State


class InitHandshakeHandler(BaseHandler):
    """
    Implements the handshake protocol after WebSocket connection is established.

    The handshake synchronizes SDK configuration with the server.

    Handshake State Machine:
        AWAITING_HELLO (initial state)
            |
            v  [Receive GATEWAY_HELLO]
        AWAITING_SYNC_COMPLETE
            |
            v  [Send WORKER_CONNECT, wait for internal sync]
        AWAITING_READY
            |
            v  [Receive GATEWAY_CONNECTION_READY]
        COMPLETE

    Reconnection:
        On reconnect, the handshake state is reset and repeated.
    """

    _closed_event: asyncio.Event | None = None
    _send_data_task: asyncio.Task[None] | None = None
    _reconnect_task: asyncio.Task[None] | None = None

    def __init__(
        self,
        logger: types.Logger,
        state: State,
        app_configs: dict[str, AppConfig],
        env: str | None,
        instance_id: str,
        max_worker_concurrency: int | None,
    ) -> None:
        self._app_configs = app_configs
        self._env = env
        self._instance_id = instance_id
        self._logger = logger
        self._handshake_state = _HandshakeState.AWAITING_HELLO
        self._state = state
        self._max_worker_concurrency = max_worker_concurrency

    def start(self) -> types.MaybeError[None]:
        err = super().start()
        if err is not None:
            return err

        if self._reconnect_task is None:
            self._reconnect_task = asyncio.create_task(
                self._reconnect_watcher(self.closed_event)
            )
        return None

    def close(self) -> None:
        super().close()

        if self._send_data_task is not None:
            self._send_data_task.cancel()
        if self._reconnect_task is not None:
            self._reconnect_task.cancel()

    async def closed(self) -> None:
        await super().closed()
        await async_lib.cancel_and_wait(self._send_data_task)
        await async_lib.cancel_and_wait(self._reconnect_task)

    async def _reconnect_watcher(self, closed_event: asyncio.Event) -> None:
        while closed_event.is_set() is False:
            await self._state.conn_init.wait_for(None, immediate=False)

            # Reset the handshake state so that we redo initialization.
            self._handshake_state = _HandshakeState.AWAITING_HELLO

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        if self._handshake_state == _HandshakeState.AWAITING_HELLO:
            if msg.kind != connect_pb2.GatewayMessageType.GATEWAY_HELLO:
                self._logger.error("Expected GATEWAY_HELLO")
                return

            self._logger.debug(
                "Handshake: AWAITING_HELLO -> AWAITING_SYNC_COMPLETE"
            )
            self._handshake_state = _HandshakeState.AWAITING_SYNC_COMPLETE

            # Reset because we were told to redo the initial handshake
            self._state.init_handshake_complete.value = False

        if self._handshake_state == _HandshakeState.AWAITING_SYNC_COMPLETE:
            self._logger.debug("Syncing")

            if (
                self._send_data_task is not None
                and not self._send_data_task.done()
            ):
                self._logger.debug("Replacing existing sync task")
                self._send_data_task.cancel()

            self._send_data_task = asyncio.create_task(
                self._send_response(
                    auth_data,
                    connection_id,
                )
            )
            return

        if self._handshake_state == _HandshakeState.AWAITING_READY:
            if (
                msg.kind
                != connect_pb2.GatewayMessageType.GATEWAY_CONNECTION_READY
            ):
                self._logger.error("Expected GATEWAY_CONNECTION_READY")
                return

            req_data = pb_utils.safe_parse(
                connect_pb2.GatewayConnectionReadyData, msg.payload
            )
            if isinstance(req_data, Exception):
                self._logger.error(
                    "Failed to parse connection ready data",
                    extra={"error": str(req_data)},
                )
                return

            extend_lease_interval = _duration_str_to_sec(
                req_data.extend_lease_interval
            )
            if isinstance(extend_lease_interval, Exception):
                self._logger.error(
                    "Failed to parse extend_lease_interval",
                )
            else:
                self._state.extend_lease_interval.value = extend_lease_interval
                self._logger.debug(
                    "Set extend lease interval",
                    extra={"value": extend_lease_interval},
                )

            self._logger.debug("Handshake: AWAITING_READY -> COMPLETE")
            self._handshake_state = _HandshakeState.COMPLETE
            self._state.conn_state.value = ConnectionState.ACTIVE
            self._state.init_handshake_complete.value = True

        return

    async def _send_response(
        self,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        ws = await self._state.ws.wait_for_not_none()

        sync_message = _create_sync_message(
            app_configs=self._app_configs,
            auth_data=auth_data,
            connection_id=connection_id,
            env=self._env,
            instance_id=self._instance_id,
            max_worker_concurrency=self._max_worker_concurrency,
        )
        if isinstance(sync_message, Exception):
            self._logger.error(
                "Failed to create sync message",
                extra={"error": str(sync_message)},
            )
            return

        err = await ws_utils.safe_send(
            self._logger,
            self._state,
            ws,
            sync_message.SerializeToString(),
        )
        if err is not None:
            self._logger.error(
                "Failed to send sync message",
                extra={"error": str(err)},
            )
            return

        self._logger.debug(
            "Handshake: AWAITING_SYNC_COMPLETE -> AWAITING_READY"
        )
        self._handshake_state = _HandshakeState.AWAITING_READY
        return None


class _HandshakeState(enum.Enum):
    """
    Explicit state machine for the WebSocket handshake protocol.

    State transitions:
        AWAITING_HELLO -> AWAITING_SYNC_COMPLETE  (on receiving GATEWAY_HELLO)
        AWAITING_SYNC_COMPLETE -> AWAITING_READY  (after sending WORKER_CONNECT)
        AWAITING_READY -> COMPLETE                (on receiving GATEWAY_CONNECTION_READY)
    """

    AWAITING_HELLO = enum.auto()
    AWAITING_SYNC_COMPLETE = enum.auto()
    AWAITING_READY = enum.auto()
    COMPLETE = enum.auto()


def _create_sync_message(
    *,
    app_configs: dict[str, AppConfig],
    auth_data: connect_pb2.AuthData,
    connection_id: str,
    env: str | None,
    instance_id: str,
    max_worker_concurrency: int | None,
) -> types.MaybeError[connect_pb2.ConnectMessage]:
    apps: list[connect_pb2.AppConfiguration] = []
    for app_id, config in app_configs.items():
        try:
            functions_bytes = pydantic_core.to_json(
                config.functions, exclude_none=True
            )
        except Exception as err:
            return err

        apps.append(
            connect_pb2.AppConfiguration(
                app_name=app_id,
                app_version=config.version,
                functions=functions_bytes,
            )
        )

    capabilities = server_lib.Capabilities().model_dump_json().encode("utf-8")

    payload = connect_pb2.WorkerConnectRequestData(
        apps=apps,
        auth_data=auth_data,
        capabilities=capabilities,
        connection_id=connection_id,
        environment=env,
        framework=server_lib.Framework.CONNECT.value,
        instance_id=instance_id,
        sdk_language=const.LANGUAGE,
        sdk_version=const.VERSION,
        system_attributes=connect_pb2.SystemAttributes(
            cpu_cores=psutil.cpu_count(),
            mem_bytes=psutil.virtual_memory().total,
            os=platform.system().lower(),
        ),
        worker_manual_readiness_ack=False,
        max_worker_concurrency=max_worker_concurrency,
    )

    return connect_pb2.ConnectMessage(
        kind=connect_pb2.GatewayMessageType.WORKER_CONNECT,
        payload=payload.SerializeToString(),
    )


def _duration_str_to_sec(duration_str: str) -> types.MaybeError[int]:
    """
    Convert a duration string (e.g. "10s") to a number of seconds. Does not
    support any other units (e.g. "1m" will error).
    """

    regex = r"^(\d+)s$"
    match = re.match(regex, duration_str)
    if match is None:
        return errors.UnreachableError(
            f"Invalid duration string: {duration_str}"
        )

    try:
        return int(match.group(1))
    except Exception:
        return errors.UnreachableError(
            f"Invalid duration string: {duration_str}"
        )



# ruff: noqa: S603, S607, T201

import os
import pathlib
import subprocess
import threading
import time
import typing

import httpx

from .command_runner import _CommandRunner

_dev_server_version: typing.Final = os.getenv("DEV_SERVER_VERSION", "latest")


class _Server:
    @property
    def origin(self) -> str:
        return f"http://0.0.0.0:{self.port}"

    def __init__(self) -> None:
        port: int
        dev_server_port_env_var = os.getenv("DEV_SERVER_PORT")
        if dev_server_port_env_var:
            port = int(dev_server_port_env_var)
        else:
            port = 8288
        self.port = port

        log_path: pathlib.Path | None = None
        if os.getenv("DEV_SERVER_LOGS") == "1":
            artifacts_dir = pathlib.Path("artifacts").absolute()
            print(f"Using artifacts directory: {artifacts_dir}")

            # Create artifacts directory if it doesn't exist.
            artifacts_dir.mkdir(exist_ok=True)

            log_path = artifacts_dir / "dev_server.log"

        self._runner = _CommandRunner(
            f"npx --ignore-scripts=false --yes inngest-cli@{_dev_server_version} dev --no-discovery --no-poll --port {self.port}",
            log_path=log_path,
        )

        self._enabled = os.getenv("DEV_SERVER_ENABLED") != "0"
        self._output_thread: threading.Thread | None = None

        self._process: subprocess.Popen[str] | None = None
        self._ready_event = threading.Event()
        self._stop_event = threading.Event()

    def start(self) -> None:
        if self._enabled is False:
            return

        print("Dev Server: starting")

        # Print inngest-cli version
        try:
            result = subprocess.run(
                [
                    "npx",
                    "--ignore-scripts=false",
                    "--yes",
                    f"inngest-cli@{_dev_server_version}",
                    "version",
                ],
                capture_output=True,
                text=True,
                timeout=10,
            )
            if result.returncode == 0:
                print(f"inngest-cli version: {result.stdout.strip()}")
            else:
                print(
                    f"Failed to get inngest-cli version: {result.stderr.strip()}"
                )
        except Exception as e:
            print(f"Error getting inngest-cli version: {e}")

        if self._runner.is_running():
            raise Exception("Dev Server is already running")

        self._runner.run()
        self._wait_for_server()
        self._runner.stop_printing()

    def _wait_for_server(self) -> None:
        print("Dev Server: waiting for start")

        while not self._ready_event.is_set():
            try:
                res = httpx.post(
                    f"http://127.0.0.1:{self.port}/v0/connect/start"
                )
                if res.status_code == 200:
                    self._ready_event.set()
                    break
            except Exception:
                time.sleep(0.1)

        print("Dev Server: started")

    def stop(self) -> None:
        if self._enabled is False:
            return

        print("Dev Server: stopping")

        if not self._runner.is_running():
            raise Exception("Dev Server is not running")

        self._runner.kill()
        print("Dev Server: stopped")


server = _Server()



from __future__ import annotations

import asyncio
import concurrent.futures
import typing

from inngest._internal import types


async def cancel_and_wait(task: asyncio.Task[typing.Any] | None) -> None:
    """
    Cancel a task and wait for its cancellation to complete.
    """

    if task is None:
        return

    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


def run_sync(
    coro: typing.Coroutine[typing.Any, typing.Any, types.T],
) -> types.MaybeError[concurrent.futures.Future[types.T]]:
    """
    Schedule an async coroutine to run in the event loop from a sync context.
    This is useful when you need to call async code from a sync function.
    """

    try:
        loop = asyncio.get_running_loop()
        return asyncio.run_coroutine_threadsafe(coro, loop)
    except RuntimeError as e:
        return e



from __future__ import annotations

import contextlib
import contextvars
import functools
import typing

from inngest._internal import step_lib

_TParams = typing.ParamSpec("_TParams")
_TReturn = typing.TypeVar("_TReturn")

_step_context_var = contextvars.ContextVar[step_lib.Step | step_lib.StepSync](
    "step"
)


def _is_step_context_set() -> bool:
    try:
        _step_context_var.get()
        return True
    except LookupError:
        return False


def get_step_context() -> step_lib.Step | step_lib.StepSync:
    return _step_context_var.get()


@contextlib.contextmanager
def set_step_context(
    step: step_lib.Step | step_lib.StepSync,
) -> typing.Generator[None, None, None]:
    if _is_step_context_set():
        raise Exception("Step context already set")

    token = _step_context_var.set(step)
    try:
        yield
    finally:
        _step_context_var.reset(token)


def step(
    step_id: str,
) -> typing.Callable[
    [typing.Callable[_TParams, _TReturn]], typing.Callable[_TParams, _TReturn]
]:
    """
    A decorator that turns a normal function into a step function. When the
    decorated function is called within an Inngest function, it will be wrapped
    with a `step.run`. When the decorated function is called outside of an
    Inngest function, it will be called directly (i.e. not wrapped with
    `step.run`)
    """

    def decorator(
        func: typing.Callable[_TParams, _TReturn],
    ) -> typing.Callable[_TParams, _TReturn]:
        @functools.wraps(func)
        def wrapper(
            *args: _TParams.args, **kwargs: _TParams.kwargs
        ) -> _TReturn:
            handler = functools.partial(func, *args, **kwargs)

            if _is_step_context_set() is False:
                return handler()

            step = _step_context_var.get()
            output = typing.cast(_TReturn, step.run(step_id, handler))  # type: ignore[arg-type]
            return output

        return wrapper

    return decorator



from __future__ import annotations

import http
import typing

from inngest._internal import net, server_lib, types

from .models import CommRequest, CommResponse

if typing.TYPE_CHECKING:
    from .handler import CommHandler


class _QueryParams(types.BaseModel):
    fn_id: str | None
    probe: server_lib.Probe | None
    step_id: str | None
    sync_id: str | None


def parse_query_params(
    query_params: dict[str, str] | dict[str, list[str]],
) -> _QueryParams | Exception:
    normalized: dict[str, str] = {}
    for k, v in query_params.items():
        if isinstance(v, list):
            normalized[k] = v[0]
        else:
            normalized[k] = v

    probe: server_lib.Probe | None = None
    probe_str = normalized.get(server_lib.QueryParamKey.PROBE.value)
    if probe_str:
        try:
            probe = server_lib.Probe(probe_str)
        except ValueError:
            return Exception(f"unsupported probe: {probe_str}")

    step_id = normalized.get(server_lib.QueryParamKey.STEP_ID.value)
    if step_id == server_lib.UNSPECIFIED_STEP_ID:
        step_id = None

    return _QueryParams(
        fn_id=normalized.get(server_lib.QueryParamKey.FUNCTION_ID.value),
        probe=probe,
        step_id=step_id,
        sync_id=normalized.get(server_lib.QueryParamKey.SYNC_ID.value),
    )


_MethodHandler = typing.Callable[
    [typing.Any, CommRequest, types.MaybeError[str | None]],
    typing.Awaitable[CommResponse | Exception],
]


def wrap_handler(
    require_signature: bool = True,
) -> typing.Callable[
    [_MethodHandler],
    typing.Callable[[typing.Any, CommRequest], typing.Awaitable[CommResponse]],
]:
    def decorator(
        method: _MethodHandler,
    ) -> typing.Callable[
        [typing.Any, CommRequest], typing.Awaitable[CommResponse]
    ]:
        """
        Perform request signature validation, error handling, header setting, and
        response signing.
        """

        async def wrapper(
            self: CommHandler,
            req: CommRequest,
        ) -> CommResponse:
            request_signing_key = None
            with req.timings.comm_handler, req.timings.async_block:
                if req.is_connect is False:
                    # Connect uses WebSockets so there isn't a request signature
                    # header.

                    req.headers = net.normalize_headers(req.headers)

                    request_signing_key = net.validate_request_sig(
                        body=req.body,
                        headers=req.headers,
                        mode=self._client._mode,
                        signing_key=self._signing_key,
                        signing_key_fallback=self._signing_key_fallback,
                    )
                    if (
                        isinstance(request_signing_key, Exception)
                        and require_signature
                    ):
                        return CommResponse.from_error(
                            self._client.logger,
                            request_signing_key,
                            status=http.HTTPStatus.UNAUTHORIZED,
                        )

                res = await method(
                    self,
                    req,
                    request_signing_key,
                )
                if isinstance(res, Exception):
                    res = CommResponse.from_error(self._client.logger, res)

            res.headers = {
                **res.headers,
                **net.create_headers(
                    env=self._client.env,
                    framework=self._framework,
                    server_kind=self._client._mode,
                ),
                server_lib.HeaderKey.SERVER_TIMING.value: req.timings.to_header(),
            }

            if isinstance(request_signing_key, str):
                err = res.sign(request_signing_key)
                if err is not None:
                    self._client.logger.error(err)

            return res

        return wrapper

    return decorator


_MethodHandlerSync = typing.Callable[
    [typing.Any, CommRequest, types.MaybeError[str | None]],
    CommResponse | Exception,
]


def wrap_handler_sync(
    require_signature: bool = True,
) -> typing.Callable[
    [_MethodHandlerSync],
    typing.Callable[[typing.Any, CommRequest], CommResponse],
]:
    def decorator(
        method: _MethodHandlerSync,
    ) -> typing.Callable[[typing.Any, CommRequest], CommResponse]:
        """
        Perform request signature validation, error handling, header setting, and
        response signing.
        """

        def wrapper(
            self: CommHandler,
            req: CommRequest,
        ) -> CommResponse:
            request_signing_key = None
            with req.timings.comm_handler:
                if req.is_connect is False:
                    # Connect uses WebSockets so there isn't a request signature
                    # header.

                    req.headers = net.normalize_headers(req.headers)

                    request_signing_key = net.validate_request_sig(
                        body=req.body,
                        headers=req.headers,
                        mode=self._client._mode,
                        signing_key=self._signing_key,
                        signing_key_fallback=self._signing_key_fallback,
                    )
                    if (
                        isinstance(request_signing_key, Exception)
                        and require_signature
                    ):
                        return CommResponse.from_error(
                            self._client.logger,
                            request_signing_key,
                            status=http.HTTPStatus.UNAUTHORIZED,
                        )

                res = method(
                    self,
                    req,
                    request_signing_key,
                )
                if isinstance(res, Exception):
                    res = CommResponse.from_error(self._client.logger, res)

            res.headers = {
                **res.headers,
                **net.create_headers(
                    env=self._client.env,
                    framework=self._framework,
                    server_kind=self._client._mode,
                ),
                server_lib.HeaderKey.SERVER_TIMING.value: req.timings.to_header(),
            }

            if isinstance(request_signing_key, str):
                err = res.sign(request_signing_key)
                if err is not None:
                    self._client.logger.error(err)

            return res

        return wrapper

    return decorator



import pydantic
import pytest

from .event import Event


def test_data_missing() -> None:
    Event(name="foo")


def test_data_list() -> None:
    with pytest.raises(pydantic.ValidationError):
        Event(
            data=[1, 2, 3],  # type: ignore[arg-type]
            name="foo",
        )


def test_data_nested() -> None:
    Event(
        data={
            "foo": {
                "bar": {
                    "baz": 1,
                },
            },
        },
        name="foo",
    )


def test_data_invalid() -> None:
    with pytest.raises(pydantic.ValidationError):
        Event(
            data=[  # type: ignore[arg-type]
                {"foo": 1},
            ],
            name="foo",
        )



"""
Remote state middleware for Inngest. This middleware allows you to store state
where you want, rather than in Inngest's infrastructure. This is useful for:
- Reducing bandwidth to/from the Inngest server.
- Avoiding step output size limits.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to use it, we recommend copying this package into your source code.
"""

from .in_memory_driver import InMemoryDriver
from .middleware import RemoteStateMiddleware, StateDriver
from .s3_driver import S3Driver

__all__ = [
    "InMemoryDriver",
    "RemoteStateMiddleware",
    "S3Driver",
    "StateDriver",
]



from __future__ import annotations

import typing

from google.protobuf import message as pb_message

from inngest._internal import types

_T = typing.TypeVar("_T", bound=pb_message.Message)


def safe_parse(pb_class: type[_T], data: bytes) -> types.MaybeError[_T]:
    """
    Parse a protobuf message, returning an Exception on failure.
    """

    try:
        msg = pb_class()
        msg.ParseFromString(data)
        return msg
    except Exception as e:
        return e



import asyncio

from inngest._internal import types

from . import connect_pb2
from .base_handler import BaseHandler
from .models import State


class DrainHandler(BaseHandler):
    """
    Handles the GATEWAY_CLOSING message for graceful gateway shutdown.

    The server sends GATEWAY_CLOSING when the gateway is shutting down (e.g.
    Inngest is doing a deploy on their end).

    When GATEWAY_CLOSING is received:
        1. The WebSocket connection is cleared via state.close_ws()
        2. This triggers the reconnection logic elsewhere
        3. The SDK connects to a different gateway

    This allows the server to gracefully drain connections without
    interrupting in-flight work.
    """

    _closed_event: asyncio.Event | None = None

    def __init__(
        self,
        logger: types.Logger,
        state: State,
    ) -> None:
        self._logger = logger
        self._state = state

    def handle_msg(
        self,
        msg: connect_pb2.ConnectMessage,
        auth_data: connect_pb2.AuthData,
        connection_id: str,
    ) -> None:
        if msg.kind != connect_pb2.GatewayMessageType.GATEWAY_CLOSING:
            return

        self._logger.debug("Draining")

        # Clear the connection to trigger reconnection logic elsewhere
        self._state.close_ws()



from __future__ import annotations

import dataclasses
import inspect
import typing
import urllib.parse

import pydantic

from inngest._internal import (
    client_lib,
    errors,
    execution_lib,
    middleware_lib,
    server_lib,
    types,
)


@dataclasses.dataclass
class _Config:
    # The user-defined function
    main: server_lib.FunctionConfig

    # The internal on_failure function
    on_failure: server_lib.FunctionConfig | None


class FunctionOpts(types.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    batch_events: server_lib.Batch | None
    cancel: list[server_lib.Cancel] | None
    concurrency: list[server_lib.Concurrency] | None
    debounce: server_lib.Debounce | None

    # Unique within an environment
    fully_qualified_id: str

    idempotency: str | None

    # Unique within an app
    local_id: str

    name: str
    on_failure: (
        execution_lib.FunctionHandlerAsync[typing.Any]
        | execution_lib.FunctionHandlerSync[typing.Any]
        | None
    )
    priority: server_lib.Priority | None
    rate_limit: server_lib.RateLimit | None
    retries: int | None
    throttle: server_lib.Throttle | None
    timeouts: server_lib.Timeouts | None
    singleton: server_lib.Singleton | None

    def convert_validation_error(
        self,
        err: pydantic.ValidationError,
    ) -> BaseException:
        return errors.FunctionConfigInvalidError.from_validation_error(err)


class Function(typing.Generic[types.T]):
    _handler: (
        execution_lib.FunctionHandlerAsync[types.T]
        | execution_lib.FunctionHandlerSync[types.T]
    )
    _on_failure_fn_id: str | None = None
    _opts: FunctionOpts
    _triggers: list[server_lib.TriggerCron | server_lib.TriggerEvent]

    @property
    def id(self) -> str:
        return self._opts.fully_qualified_id

    @property
    def is_handler_async(self) -> bool:
        """Whether the main handler is async."""
        return inspect.iscoroutinefunction(self._handler)

    @property
    def is_on_failure_handler_async(self) -> bool | None:
        """
        Whether the on_failure handler is async. Returns None if there isn't an
        on_failure handler.
        """
        if self._opts.on_failure is None:
            return None
        return inspect.iscoroutinefunction(self._opts.on_failure)

    @property
    def local_id(self) -> str:
        return self._opts.local_id

    @property
    def name(self) -> str:
        return self._opts.name

    @property
    def on_failure_fn_id(self) -> str | None:
        return self._on_failure_fn_id

    def __init__(
        self,
        opts: FunctionOpts,
        trigger: server_lib.TriggerCron
        | server_lib.TriggerEvent
        | list[server_lib.TriggerCron | server_lib.TriggerEvent],
        handler: execution_lib.FunctionHandlerAsync[types.T]
        | execution_lib.FunctionHandlerSync[types.T],
        output_type: object = types.EmptySentinel,
        middleware: list[middleware_lib.UninitializedMiddleware] | None = None,
    ) -> None:
        self._handler = handler
        self._middleware = middleware or []
        self._opts = opts
        self._output_type = output_type
        self._triggers = trigger if isinstance(trigger, list) else [trigger]

        if opts.on_failure is not None:
            if (
                self.is_handler_async
                and self.is_on_failure_handler_async is False
            ):
                raise errors.Error(
                    f"an async function cannot have a non-async on_failure handler (function {opts.local_id})"
                )
            if (
                self.is_handler_async is False
                and self.is_on_failure_handler_async is True
            ):
                raise errors.Error(
                    f"a non-async function cannot have an async on_failure handler (function {opts.local_id})"
                )

            self._on_failure_fn_id = f"{opts.fully_qualified_id}-failure"

    async def call(
        self,
        client: client_lib.Inngest,
        ctx: execution_lib.Context,
        fn_id: str,
        middleware: middleware_lib.MiddlewareManager,
    ) -> execution_lib.CallResult:
        for m in self._middleware:
            middleware.add(m)

        if self.id == fn_id:
            handler = self._handler
            output_type = self._output_type
        elif self.on_failure_fn_id == fn_id:
            if self._opts.on_failure is None:
                return execution_lib.CallResult(
                    errors.FunctionNotFoundError("on_failure not defined")
                )
            handler = self._opts.on_failure

            # We only need to serialize to JSON so any type is fine.
            # Deserialization isn't necessary since on_failure handlers aren't
            # invoked via `step.invoke`.
            output_type = object
        else:
            return execution_lib.CallResult(
                errors.FunctionNotFoundError("function ID mismatch")
            )

        if not execution_lib.is_function_handler_async(handler):
            raise errors.UnreachableError("handler is not async")

        with execution_lib.set_step_context(ctx.step):
            call_res = await ctx.step._execution.run(
                client,
                ctx,
                handler,
                self,
                output_type,
            )

        err = await middleware.transform_output(call_res)
        if isinstance(err, Exception):
            return execution_lib.CallResult(err)

        err = await middleware.before_response()
        if isinstance(err, Exception):
            return execution_lib.CallResult(err)

        return call_res

    def call_sync(
        self,
        client: client_lib.Inngest,
        ctx: execution_lib.ContextSync,
        fn_id: str,
        middleware: middleware_lib.MiddlewareManager,
    ) -> execution_lib.CallResult:
        for m in self._middleware:
            middleware.add(m)

        if self.id == fn_id:
            handler = self._handler
            output_type = self._output_type
        elif self.on_failure_fn_id == fn_id:
            if self._opts.on_failure is None:
                return execution_lib.CallResult(
                    errors.FunctionNotFoundError("on_failure not defined")
                )
            handler = self._opts.on_failure

            # We only need to serialize to JSON so any type is fine.
            # Deserialization isn't necessary since on_failure handlers aren't
            # invoked via `step.invoke`.
            output_type = object
        else:
            return execution_lib.CallResult(
                errors.FunctionNotFoundError("function ID mismatch")
            )

        if not execution_lib.is_function_handler_sync(handler):
            raise errors.UnreachableError("handler is not sync")

        with execution_lib.set_step_context(ctx.step):
            # We don't need to pass a thread pool here because the sync handler
            # is not used by Connect.
            call_res = ctx.step._execution.run(
                client,
                ctx,
                handler,
                self,
                output_type,
            )

        err = middleware.transform_output_sync(call_res)
        if isinstance(err, Exception):
            return execution_lib.CallResult(err)

        err = middleware.before_response_sync()
        if isinstance(err, Exception):
            return execution_lib.CallResult(err)

        return call_res

    def get_config(self, app_url: str) -> _Config:
        fn_id = self._opts.fully_qualified_id
        name = self._opts.name

        if self._opts.retries is not None:
            retries = server_lib.Retries(attempts=self._opts.retries)
        else:
            retries = None

        url = (
            app_url
            + "?"
            + urllib.parse.urlencode(
                {
                    "fnId": fn_id,
                    "stepId": server_lib.ROOT_STEP_ID,
                }
            )
        )

        runtime_type: typing.Literal["http", "ws"] = "http"
        if url.startswith("wss://"):
            runtime_type = "ws"

        main = server_lib.FunctionConfig(
            batch_events=self._opts.batch_events,
            cancel=self._opts.cancel,
            concurrency=self._opts.concurrency,
            debounce=self._opts.debounce,
            id=fn_id,
            idempotency=self._opts.idempotency,
            name=name,
            priority=self._opts.priority,
            rate_limit=self._opts.rate_limit,
            steps={
                server_lib.ROOT_STEP_ID: server_lib.Step(
                    id=server_lib.ROOT_STEP_ID,
                    name=server_lib.ROOT_STEP_ID,
                    retries=retries,
                    runtime=server_lib.Runtime(
                        type=runtime_type,
                        url=url,
                    ),
                ),
            },
            throttle=self._opts.throttle,
            timeouts=self._opts.timeouts,
            triggers=self._triggers,
            singleton=self._opts.singleton,
        )

        on_failure = None
        if self.on_failure_fn_id is not None:
            url = (
                app_url
                + "?"
                + urllib.parse.urlencode(
                    {
                        "fnId": self.on_failure_fn_id,
                        "stepId": server_lib.ROOT_STEP_ID,
                    }
                )
            )

            on_failure = server_lib.FunctionConfig(
                batch_events=None,
                cancel=[],
                concurrency=[],
                debounce=None,
                id=self.on_failure_fn_id,
                idempotency=None,
                name=f"{name} (failure)",
                priority=None,
                rate_limit=None,
                steps={
                    server_lib.ROOT_STEP_ID: server_lib.Step(
                        id=server_lib.ROOT_STEP_ID,
                        name=server_lib.ROOT_STEP_ID,
                        retries=server_lib.Retries(attempts=0),
                        runtime=server_lib.Runtime(
                            type="http",
                            url=url,
                        ),
                    )
                },
                throttle=None,
                timeouts=None,
                triggers=[
                    server_lib.TriggerEvent(
                        event=server_lib.InternalEvents.FUNCTION_FAILED.value,
                        expression=f"event.data.function_id == '{self.id}'",
                    )
                ],
                singleton=None,
            )

        return _Config(main=main, on_failure=on_failure)

    def get_id(self) -> str:
        return self._opts.fully_qualified_id



from __future__ import annotations

import json
import secrets
import string
import typing

import pydantic
import typing_extensions

import inngest

from .middleware import StateDriver

if typing.TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client


class _StateSurrogate(pydantic.BaseModel):
    """
    Replaces step output sent back to Inngest. Its data is sufficient to
    retrieve the actual state.
    """

    bucket: str
    key: str


class S3Driver(StateDriver):
    """
    S3 driver for remote state middleware.
    """

    # Marker to indicate that the data is stored remotely.
    _marker: typing.Final = "__REMOTE_STATE__"

    # Marker to indicate which strategy was used. This is useful for knowing
    # whether the official S3 driver was used.
    _strategy_marker: typing.Final = "__STRATEGY__"

    _strategy_identifier: typing.Final = "inngest/s3"

    def __init__(
        self,
        *,
        bucket: str,
        client: S3Client,
    ) -> None:
        """
        Args:
        ----
            bucket: Bucket name to store remote state.
            client: Boto3 S3 client.
        """

        self._bucket = bucket
        self._client = client

    def _create_key(self) -> str:
        chars = string.ascii_letters + string.digits
        return "".join(secrets.choice(chars) for _ in range(32))

    def _is_remote(
        self, data: object
    ) -> typing_extensions.TypeGuard[dict[str, object]]:
        return (
            isinstance(data, dict)
            and self._marker in data
            and self._strategy_marker in data
            and data[self._strategy_marker] == self._strategy_identifier
        )

    def load_steps(self, steps: inngest.StepMemos) -> None:
        """
        Hydrate steps with remote state if necessary.

        Args:
        ----
            steps: Steps that may need hydration.
        """

        for step in steps.values():
            if not self._is_remote(step.data):
                continue

            surrogate = _StateSurrogate.model_validate(step.data)

            step.data = json.loads(
                self._client.get_object(
                    Bucket=surrogate.bucket,
                    Key=surrogate.key,
                )["Body"]
                .read()
                .decode()
            )

    def save_step(
        self,
        run_id: str,
        value: object,
    ) -> dict[str, object]:
        """
        Save a step's output to the remote store and return a placeholder.

        Args:
        ----
            run_id: Run ID.
            value: Step output.
        """

        key = f"inngest/remote_state/{run_id}/{self._create_key()}"
        self._client.put_object(
            Body=json.dumps(value),
            Bucket=self._bucket,
            Key=key,
        )

        surrogate = {
            self._marker: True,
            self._strategy_marker: self._strategy_identifier,
            **_StateSurrogate(bucket=self._bucket, key=key).model_dump(),
        }

        return surrogate



from __future__ import annotations

import dataclasses
import enum

import websockets

from inngest._internal import server_lib

from . import connect_pb2
from .value_watcher import ValueWatcher


@dataclasses.dataclass
class AppConfig:
    functions: list[server_lib.FunctionConfig]
    version: str | None


@dataclasses.dataclass
class State:
    """
    Shared state for Connect.
    """

    conn_id: ValueWatcher[str | None]
    conn_init: ValueWatcher[tuple[connect_pb2.AuthData, str] | None]
    conn_state: ValueWatcher[ConnectionState]
    exclude_gateways: ValueWatcher[list[str]]
    extend_lease_interval: ValueWatcher[int | None]

    # Error that should make Connect close and raise an exception.
    fatal_error: ValueWatcher[Exception | None]

    init_handshake_complete: ValueWatcher[bool]

    # Number of pending requests. This is useful for handlers that need to wait
    # for all pending requests to complete before closing.
    pending_request_count: ValueWatcher[int]

    ws: ValueWatcher[websockets.ClientConnection | None]

    def allow_reconnect(self) -> bool:
        return self.conn_state.value not in [
            ConnectionState.CLOSED,
            ConnectionState.CLOSING,
        ]

    def close_ws(self) -> None:
        """
        Close the WebSocket connection
        """

        if self.allow_reconnect():
            self.conn_state.value = ConnectionState.RECONNECTING
        self.conn_init.value = None
        self.ws.value = None


class ConnectionState(enum.Enum):
    """
    State of the connection to the Inngest server.
    """

    ACTIVE = "ACTIVE"
    CLOSED = "CLOSED"
    CLOSING = "CLOSING"
    CONNECTING = "CONNECTING"
    PAUSED = "PAUSED"
    RECONNECTING = "RECONNECTING"



import typing


class BaseAdapter(typing.Protocol):
    """
    Base AI adapter.
    """

    def auth_key(self) -> str:
        """
        Return the authentication key for the adapter.
        """
        ...

    def format(self) -> str:
        """
        Return the format for the adapter.
        """
        ...

    def headers(self) -> dict[str, str]:
        """
        Return the headers for the adapter.
        """
        ...

    def on_call(self, body: dict[str, object]) -> None:
        """
        Modify the request body.
        """
        ...

    def url_infer(self) -> str:
        """
        Return the URL for generating text.
        """
        ...



import asyncio
import threading
import typing
import unittest

import pytest

from .value_watcher import ValueWatcher


class TestValueWatcher(unittest.IsolatedAsyncioTestCase):
    async def _start_tasks(
        self, *coros: typing.Coroutine[typing.Any, typing.Any, typing.Any]
    ) -> typing.Callable[[], typing.Awaitable[None]]:
        tasks = [asyncio.create_task(coro) for coro in coros]
        for task in tasks:
            self.addCleanup(task.cancel)

        # Allow an event loop iteration to pass, ensuring that the tasks start.
        await asyncio.sleep(0)

        async def wait() -> None:
            await asyncio.gather(*tasks)

        return wait

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for(self) -> None:
        watcher = ValueWatcher(0)

        # Immediately returns because the value is already 0.
        result = await watcher.wait_for(0)
        self.assertEqual(result, 0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1),
        )

        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_not(self) -> None:
        watcher = ValueWatcher(0)

        # Immediately returns because the value is already 0.
        result = await watcher.wait_for_not(1)
        self.assertEqual(result, 0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for_not(0),
        )

        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_change(self) -> None:
        watcher = ValueWatcher(0)

        task = asyncio.create_task(watcher.wait_for_change())
        self.addCleanup(task.cancel)
        await asyncio.sleep(0)

        watcher.value = 1

        result = await task
        self.assertEqual(result, 1)

    @pytest.mark.timeout(2, method="thread")
    async def test_watch_wait_for_and_not(self) -> None:
        """
        Match parallel watchers:
        - Wait for the value to not be a certain value.
        - Wait for the value to be a certain value.
        """

        watcher = ValueWatcher(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for_not(0),
            watcher.wait_for(1),
        )

        # Set the value, which should propagate to all watchers.
        watcher.value = 1

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_quickly_set_twice(self) -> None:
        """
        Setting twice really quickly should not cause a missed value.
        """

        watcher = ValueWatcher(0)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1),
            watcher.wait_for(2),
        )

        # Set the value, which should propagate to all watchers.
        watcher.value = 1
        watcher.value = 2

        # Wait for all watchers to complete.
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_immediate_false_waits_even_if_already_matching(self) -> None:
        """
        immediate=False should not return for a pre-existing match. It must
        wait for an actual transition.
        """

        watcher = ValueWatcher(1)

        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1, immediate=False),
        )

        # Value is already 1, but immediate=False means we're waiting for the
        # next transition to 1. Change away and back.
        watcher.value = 0
        watcher.value = 1

        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_timeout_raises(self) -> None:
        watcher = ValueWatcher(0)

        with self.assertRaises(asyncio.TimeoutError):
            await watcher.wait_for(1, timeout=0.05)

    @pytest.mark.timeout(2, method="thread")
    async def test_timeout_raises_wait_for_not(self) -> None:
        watcher = ValueWatcher(0)

        with self.assertRaises(asyncio.TimeoutError):
            await watcher.wait_for_not(0, timeout=0.05)

    @pytest.mark.timeout(2, method="thread")
    async def test_timeout_raises_wait_for_not_none(self) -> None:
        watcher = ValueWatcher[str | None](None)

        with self.assertRaises(asyncio.TimeoutError):
            await watcher.wait_for_not_none(timeout=0.05)

    @pytest.mark.timeout(2, method="thread")
    async def test_timeout_raises_wait_for_change(self) -> None:
        watcher = ValueWatcher(0)

        with self.assertRaises(asyncio.TimeoutError):
            await watcher.wait_for_change(timeout=0.05)

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_not_none(self) -> None:
        watcher = ValueWatcher[str | None](None)

        task = asyncio.create_task(watcher.wait_for_not_none())
        self.addCleanup(task.cancel)
        await asyncio.sleep(0)

        watcher.value = "hello"

        result = await task
        self.assertEqual(result, "hello")

    @pytest.mark.timeout(2, method="thread")
    async def test_wait_for_not_none_immediate(self) -> None:
        watcher = ValueWatcher[str | None]("hello")

        result = await watcher.wait_for_not_none()
        self.assertEqual(result, "hello")

    @pytest.mark.timeout(2, method="thread")
    async def test_dedup_does_not_notify(self) -> None:
        """
        Setting the same value should not trigger a notification.
        """

        watcher = ValueWatcher(0)

        task = asyncio.create_task(watcher.wait_for(1))
        self.addCleanup(task.cancel)
        await asyncio.sleep(0)

        # Setting to the current value should be a no-op.
        watcher.value = 0
        watcher.value = 0

        # The specific waiter task should still be pending.
        await asyncio.sleep(0)
        self.assertFalse(task.done())

        # Now actually change it.
        watcher.value = 1
        result = await task
        self.assertEqual(result, 1)

    @pytest.mark.timeout(2, method="thread")
    async def test_on_change_callback(self) -> None:
        changes: list[tuple[int, int]] = []

        watcher = ValueWatcher(
            0, on_change=lambda old, new: changes.append((old, new))
        )
        watcher.value = 1
        watcher.value = 2

        self.assertEqual(changes, [(0, 1), (1, 2)])

    @pytest.mark.timeout(2, method="thread")
    async def test_on_change_method(self) -> None:
        """
        The on_change() instance method registers callbacks post-construction.
        """

        watcher = ValueWatcher(0)

        changes: list[tuple[int, int]] = []
        watcher.on_change(lambda old, new: changes.append((old, new)))

        watcher.value = 1
        watcher.value = 2

        self.assertEqual(changes, [(0, 1), (1, 2)])

    @pytest.mark.timeout(2, method="thread")
    async def test_on_change_exception_does_not_block_others(self) -> None:
        """
        A throwing callback must not prevent subsequent callbacks from firing.
        """

        changes: list[tuple[int, int]] = []

        def bad_callback(old: int, new: int) -> None:
            raise RuntimeError("boom")

        watcher = ValueWatcher(0, on_change=bad_callback)
        watcher.on_change(lambda old, new: changes.append((old, new)))

        watcher.value = 1

        self.assertEqual(changes, [(0, 1)])

    async def test_watch_returns_queue(self) -> None:
        """
        `_watch` must return a Queue, not an async generator. An async
        generator causes "Task was destroyed but it is pending!" warnings
        when the event loop closes shortly after `wait_for` completes.
        """

        watcher = ValueWatcher(0)
        with watcher._watch() as result:
            self.assertIsInstance(result, asyncio.Queue)

    @pytest.mark.timeout(2, method="thread")
    async def test_watch_queue_cleanup(self) -> None:
        """
        Watch queues must be removed after a waiter completes.
        """

        watcher = ValueWatcher(0)
        self.assertEqual(len(watcher._watch_queues), 0)

        task = asyncio.create_task(watcher.wait_for(1))
        self.addCleanup(task.cancel)
        await asyncio.sleep(0)

        self.assertEqual(len(watcher._watch_queues), 1)

        watcher.value = 1
        await task

        self.assertEqual(len(watcher._watch_queues), 0)

    @pytest.mark.timeout(2, method="thread")
    async def test_cross_thread_notification(self) -> None:
        """
        Value set from another thread should notify waiters.

        This test is important because different threads have different event
        loops. Some asyncio primitives are not thread-safe, so we need to make
        sure we aren't using them.
        """

        # Create a watcher in this thread
        watcher = ValueWatcher(0)

        # Create a waiter in this thread
        wait_for_tasks = await self._start_tasks(
            watcher.wait_for(1),
        )

        def other_thread() -> None:
            # Set the value in another thread
            watcher.value = 1

        thread = threading.Thread(target=other_thread)
        thread.start()
        thread.join()

        # Wait for the notification to come from the other thread
        await wait_for_tasks()

    @pytest.mark.timeout(2, method="thread")
    async def test_closed_loop_does_not_raise(self) -> None:
        """
        Setting the value after a watcher's event loop has closed must not
        raise.
        """

        watcher = ValueWatcher(0)

        # Inject a queue bound to a closed event loop.
        closed_loop = asyncio.new_event_loop()
        closed_loop.close()
        queue: asyncio.Queue[tuple[int, int]] = asyncio.Queue()
        watcher._watch_queues.append((closed_loop, queue))

        # Should not raise despite the closed loop.
        watcher.value = 1

    @pytest.mark.timeout(2, method="thread")
    async def test_on_value(self) -> None:
        watcher = ValueWatcher(0)

        called = asyncio.Event()
        watcher.on_value(1, called.set)

        watcher.value = 1
        await called.wait()



"""
Experimental entrypoint for the Inngest SDK.

Does not follow semantic versioning! Breaking changes may occur at any time.
"""

from inngest._internal.execution_lib import (
    get_step_context,
    set_step_context,
    step,
)

__all__ = ["get_step_context", "set_step_context", "step"]



"""
Django integration for Inngest.
"""

import http
import json
import typing

import django
import django.http
import django.urls
import django.views.decorators.csrf

from ._internal import (
    client_lib,
    comm_lib,
    config_lib,
    const,
    function,
    server_lib,
    transforms,
)

FRAMEWORK = server_lib.Framework.DJANGO


def serve(
    client: client_lib.Inngest,
    functions: list[function.Function[typing.Any]],
    *,
    public_path: str | None = None,
    serve_origin: str | None = None,
    serve_path: str | None = None,
) -> django.urls.URLPattern:
    """
    Serve Inngest functions in a Django app.

    Args:
    ----
        client: Inngest client.
        functions: List of functions to serve.
        public_path: Path that the Inngest server sends requests to. This is only necessary if the SDK is behind a proxy that rewrites the path.
        serve_origin: Origin for serving Inngest functions. This is typically only useful during Docker-based development.
        serve_path: Path for serving Inngest functions. This is only useful if you don't want serve Inngest at the /api/inngest path.
    """

    handler = comm_lib.CommHandler(
        client=client,
        framework=FRAMEWORK,
        functions=functions,
        streaming=const.Streaming.DISABLE,  # Not supported yet.
    )

    async_mode = any(
        function.is_handler_async or function.is_on_failure_handler_async
        for function in functions
    )

    if async_mode:
        return _create_handler_async(
            client,
            handler,
            public_path=public_path,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )
    else:
        return _create_handler_sync(
            client,
            handler,
            public_path=public_path,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )


def _create_handler_sync(
    client: client_lib.Inngest,
    handler: comm_lib.CommHandler,
    *,
    public_path: str | None,
    serve_origin: str | None,
    serve_path: str | None,
) -> django.urls.URLPattern:
    def inngest_api(
        request: django.http.HttpRequest,
    ) -> django.http.HttpResponse:
        comm_req = comm_lib.CommRequest(
            body=request.body,
            headers=dict(request.headers.items()),
            public_path=public_path,
            query_params=dict(request.GET.items()),
            raw_request=request,
            request_url=request.build_absolute_uri(),
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        if request.method == "GET":
            return _to_response(
                client,
                handler.get_sync(comm_req),
            )

        if request.method == "POST":
            return _to_response(
                client,
                handler.post_sync(comm_req),
            )

        if request.method == "PUT":
            return _to_response(
                client,
                handler.put_sync(comm_req),
            )

        return django.http.JsonResponse(
            {"error": "Unsupported method"},
            status=http.HTTPStatus.METHOD_NOT_ALLOWED,
        )

    return django.urls.path(
        _trim_leading_slash(
            config_lib.get_serve_path(serve_path) or const.DEFAULT_SERVE_PATH
        ),
        django.views.decorators.csrf.csrf_exempt(inngest_api),
    )


def _create_handler_async(
    client: client_lib.Inngest,
    handler: comm_lib.CommHandler,
    *,
    public_path: str | None,
    serve_origin: str | None,
    serve_path: str | None,
) -> django.urls.URLPattern:
    major_version = transforms.get_major_version(django.get_version())
    if isinstance(major_version, Exception):
        client.logger.error(major_version)
    else:
        if major_version < 5:
            # Django 5 introduced async support for csrf_exempt
            raise Exception(
                "Django version 5 or higher is required for async mode"
            )

    async def inngest_api(
        request: django.http.HttpRequest,
    ) -> django.http.HttpResponse:
        comm_req = comm_lib.CommRequest(
            body=request.body,
            headers=dict(request.headers.items()),
            public_path=public_path,
            query_params=dict(request.GET.items()),
            raw_request=request,
            request_url=request.build_absolute_uri(),
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        if request.method == "GET":
            return _to_response(
                client,
                handler.get_sync(comm_req),
            )

        if request.method == "POST":
            return _to_response(
                client,
                await handler.post(comm_req),
            )

        if request.method == "PUT":
            return _to_response(
                client,
                await handler.put(comm_req),
            )

        return django.http.JsonResponse(
            {"error": "Unsupported method"},
            status=http.HTTPStatus.METHOD_NOT_ALLOWED,
        )

    return django.urls.path(
        _trim_leading_slash(
            config_lib.get_serve_path(serve_path) or const.DEFAULT_SERVE_PATH
        ),
        django.views.decorators.csrf.csrf_exempt(inngest_api),
    )


def _to_response(
    client: client_lib.Inngest,
    comm_res: comm_lib.CommResponse,
) -> django.http.HttpResponse:
    body = transforms.dump_json(comm_res.body)
    if isinstance(body, Exception):
        comm_res = comm_lib.CommResponse.from_error(client.logger, body)
        body = json.dumps(comm_res.body)

    return django.http.HttpResponse(
        body.encode("utf-8"),
        headers=comm_res.headers,
        status=comm_res.status_code,
    )


def _trim_leading_slash(value: str) -> str:
    return value.lstrip("/")



import datetime
import hashlib
import inspect
import json
import re
import traceback
import typing

import jcs

from inngest._internal import errors, server_lib, types


def get_traceback(err: Exception) -> str:
    return "".join(
        traceback.format_exception(type(err), err, err.__traceback__)
    )


def hash_event_key(key: str) -> str:
    return hashlib.sha256(key.encode("utf-8")).hexdigest()


def hash_signing_key(key: str) -> str:
    return hashlib.sha256(
        bytearray.fromhex(remove_signing_key_prefix(key))
    ).hexdigest()


def hash_step_id(step_id: str) -> str:
    return hashlib.sha1(step_id.encode("utf-8")).hexdigest()  # noqa: S324


def dump_json(obj: object) -> types.MaybeError[str]:
    try:
        return json.dumps(obj)
    except Exception as err:
        return errors.OutputUnserializableError(str(err))


def canonicalize(value: bytes) -> types.MaybeError[bytes]:
    if len(value) == 0:
        return value

    try:
        loaded = json.loads(value)
        value_jcs = jcs.canonicalize(loaded)
        if not isinstance(value_jcs, bytes):
            return Exception("failed to canonicalize")
        return value_jcs
    except Exception as err:
        return Exception("failed to canonicalize: " + str(err))


def remove_signing_key_prefix(key: str) -> str:
    prefix_match = re.match(r"^signkey-[\w]+-", key)
    prefix = ""
    if prefix_match:
        prefix = prefix_match.group(0)

    return key[len(prefix) :]


def deep_strip_none(obj: types.T) -> types.T:
    """
    Recursively remove items whose value is None.
    """

    if isinstance(obj, dict):
        return {k: deep_strip_none(v) for k, v in obj.items() if v is not None}  # type: ignore
    if isinstance(obj, list):
        return [deep_strip_none(v) for v in obj if v is not None]  # type: ignore
    return obj


class _Duration:
    @classmethod
    def second(cls, count: int = 1) -> int:
        return count * 1000

    @classmethod
    def minute(cls, count: int = 1) -> int:
        return count * cls.second(60)

    @classmethod
    def hour(cls, count: int = 1) -> int:
        return count * cls.minute(60)

    @classmethod
    def day(cls, count: int = 1) -> int:
        return count * cls.hour(24)

    @classmethod
    def week(cls, count: int = 1) -> int:
        return count * cls.day(7)


def to_duration_str(
    ms: int | datetime.timedelta,
) -> types.MaybeError[str]:
    if isinstance(ms, datetime.timedelta):
        ms = int(ms.total_seconds() * 1000)

    if ms < _Duration.second():
        return errors.FunctionConfigInvalidError(
            "duration must be at least 1 second"
        )

    # Use the largest unit that divides evenly to avoid truncation
    for unit, suffix in [
        (_Duration.week(), "w"),
        (_Duration.day(), "d"),
        (_Duration.hour(), "h"),
        (_Duration.minute(), "m"),
        (_Duration.second(), "s"),
    ]:
        if ms % unit == 0:
            return f"{ms // unit}{suffix}"

    return errors.FunctionConfigInvalidError(
        "duration must be a whole number of seconds"
    )


def to_maybe_duration_str(
    ms: int | datetime.timedelta | None,
) -> types.MaybeError[str | None]:
    if ms is None:
        return None
    return to_duration_str(ms)


def to_iso_utc(value: datetime.datetime) -> str:
    return (
        value.astimezone(datetime.timezone.utc).strftime(
            "%Y-%m-%dT%H:%M:%S.%f"
        )[:-3]
        + "Z"
    )


def _to_int(value: typing.Any) -> types.MaybeError[int]:
    try:
        return int(value)
    except Exception as err:
        return ValueError(f"invalid integer: {err}")


def get_major_version(version: str) -> types.MaybeError[int]:
    return _to_int(version.split(".")[0])


def get_server_kind(
    headers: dict[str, str],
) -> server_lib.ServerKind | None | Exception:
    value = headers.get(server_lib.HeaderKey.SERVER_KIND.value, None)
    if value is None:
        return None

    try:
        return server_lib.ServerKind(value)
    except ValueError:
        return Exception(f"invalid server kind: {value}")


async def maybe_await(
    value: types.T | typing.Awaitable[types.T],
) -> types.T:
    if inspect.isawaitable(value):
        return await value  # type: ignore

    return value  # type: ignore


def remove_first_traceback_frame(err: Exception) -> None:
    """
    Remove the first frame from the traceback, since we don't want our internal
    code to appear in the traceback.
    """

    if err.__traceback__:
        err.__traceback__ = err.__traceback__.tb_next



class NonRetryableError(Exception):
    """
    An error that should not be retried.

    Retry logic should check for this error type and stop immediately rather
    than continuing to retry.
    """

    pass


class UnreachableError(Exception):
    """
    Should never be raised. If it is, it indicates a bug in the code.

    This is used for cases that should be logically impossible, such as
    exhaustive pattern matching where all cases are handled.
    """

    pass



from __future__ import annotations

import typing

import pydantic

from inngest._internal import types


class Event(types.BaseModel):
    data: typing.Mapping[str, types.JSON] = {}
    id: str = ""
    name: str
    ts: int = 0

    @pydantic.field_validator("data", mode="before")
    @classmethod
    def _ensure_dict(
        cls,
        v: typing.Mapping[str, types.JSON] | None,
    ) -> typing.Mapping[str, types.JSON]:
        """
        Necessary because sometimes event data is sent as null, but we want
        users to interact with data as a dict
        """

        return v or {}


# Necessary because of the recursive JSON type
Event.model_rebuild()
